"""Representation of coordinates in different systems."""
# ruff: noqa: N803, N806

__all__: list[str] = []

import functools as ft
from typing import Any

import equinox as eqx
import jax
from plum import dispatch

import quaxed.lax as qlax
import quaxed.numpy as jnp
import unxt as u

import coordinax._src.vectors.custom_types as ct
from .base import AbstractAcc3D, AbstractPos3D, AbstractVel3D
from .base_spherical import AbstractSphericalPos, _90d, _180d, _360d
from .cartesian import CartesianAcc3D, CartesianPos3D, CartesianVel3D
from .cylindrical import CylindricalAcc, CylindricalPos, CylindricalVel
from .generic import Cartesian3D
from .lonlatspherical import (
    LonCosLatSphericalVel,
    LonLatSphericalAcc,
    LonLatSphericalPos,
    LonLatSphericalVel,
)
from .mathspherical import MathSphericalAcc, MathSphericalPos, MathSphericalVel
from .spherical import SphericalAcc, SphericalPos, SphericalVel
from .spheroidal import ProlateSpheroidalAcc, ProlateSpheroidalPos, ProlateSpheroidalVel
from coordinax._src.vectors.base import AbstractVector
from coordinax._src.vectors.base_pos import AbstractPos
from coordinax._src.vectors.private_api import combine_aux, wrap_vconvert_impl_params

###############################################################################


@dispatch(precedence=1)
def vector(cls: type[AbstractPos3D], obj: AbstractPos3D, /) -> AbstractPos3D:
    """Construct from a 3D position.

    Examples
    --------
    >>> import coordinax.vecs as cxv

    >>> cart = cxv.CartesianPos3D.from_([1, 2, 3], "km")
    >>> cxv.AbstractPos3D.from_(cart) is cart
    True

    >>> sph = cart.vconvert(cxv.SphericalPos)
    >>> cxv.AbstractPos3D.from_(sph) is sph
    True

    >>> cyl = cart.vconvert(cxv.CylindricalPos)
    >>> cxv.AbstractPos3D.from_(cyl) is cyl
    True

    """
    return obj


# ---------------------------------------------------------


@dispatch(precedence=1)
def vector(cls: type[AbstractVel3D], obj: AbstractVel3D, /) -> AbstractVel3D:
    """Construct from a 3D velocity.

    Examples
    --------
    >>> import coordinax.vecs as cxv

    >>> q = cxv.CartesianPos3D.from_([1, 1, 1], "km")

    >>> cart = cxv.CartesianVel3D.from_([1, 2, 3], "km/s")
    >>> cxv.AbstractVel3D.from_(cart) is cart
    True

    >>> sph = cart.vconvert(cxv.SphericalVel, q)
    >>> cxv.AbstractVel3D.from_(sph) is sph
    True

    >>> cyl = cart.vconvert(cxv.CylindricalVel, q)
    >>> cxv.AbstractVel3D.from_(cyl) is cyl
    True

    """
    return obj


# ---------------------------------------------------------


@dispatch(precedence=1)
def vector(cls: type[AbstractAcc3D], obj: AbstractAcc3D, /) -> AbstractAcc3D:
    """Construct from a 3D velocity.

    Examples
    --------
    >>> import coordinax.vecs as cxv

    >>> q = cxv.CartesianPos3D.from_([1, 1, 1], "km")
    >>> p = cxv.CartesianVel3D.from_([1, 1, 1], "km/s")

    >>> cart = cxv.CartesianAcc3D.from_([1, 2, 3], "km/s2")
    >>> cxv.AbstractAcc3D.from_(cart) is cart
    True

    >>> sph = cart.vconvert(cxv.SphericalAcc, p, q)
    >>> cxv.AbstractAcc3D.from_(sph) is sph
    True

    >>> cyl = cart.vconvert(cxv.CylindricalAcc, p, q)
    >>> cxv.AbstractAcc3D.from_(cyl) is cyl
    True

    """
    return obj


# ---------------------------------------------------------


@dispatch
def vector(
    cls: type[SphericalPos],
    *,
    r: u.AbstractQuantity,
    theta: u.AbstractQuantity,
    phi: u.AbstractQuantity,
) -> SphericalPos:
    """Construct SphericalPos, allowing for out-of-range values.

    Examples
    --------
    >>> import unxt as u
    >>> import coordinax.vecs as cxv

    Let's start with a valid input:

    >>> vec = cxv.SphericalPos.from_(r=u.Quantity(3, "km"),
    ...                              theta=u.Quantity(90, "deg"),
    ...                              phi=u.Quantity(0, "deg"))
    >>> print(vec)
    <SphericalPos (r[km], theta[deg], phi[deg])
        [ 3 90  0]>

    The radial distance can be negative, which wraps the azimuthal angle by 180
    degrees and flips the polar angle:

    >>> vec = cxv.SphericalPos.from_(r=u.Quantity(-3, "km"),
    ...                              theta=u.Quantity(45, "deg"),
    ...                              phi=u.Quantity(0, "deg"))
    >>> print(vec)
    <SphericalPos (r[km], theta[deg], phi[deg])
        [  3 135 180]>

    The polar angle can be outside the [0, 180] deg range, causing the azimuthal
    angle to be shifted by 180 degrees:

    >>> vec = cxv.SphericalPos.from_(r=u.Quantity(3, "km"),
    ...                              theta=u.Quantity(190, "deg"),
    ...                              phi=u.Quantity(0, "deg"))
    >>> print(vec)
    <SphericalPos (r[km], theta[deg], phi[deg])
        [  3 170 180]>

    The azimuth can be outside the [0, 360) deg range. This is wrapped to the
    [0, 360) deg range (actually the base from_ does this):

    >>> vec = cxv.SphericalPos.from_(r=u.Quantity(3, "km"),
    ...                              theta=u.Quantity(90, "deg"),
    ...                              phi=u.Quantity(365, "deg"))
    >>> vec.phi
    Angle(Array(5, dtype=int32, ...), unit='deg')

    """
    # 1) Convert the inputs
    fields = SphericalPos.__dataclass_fields__
    r = fields["r"].metadata["converter"](r)
    theta = fields["theta"].metadata["converter"](theta)
    phi = fields["phi"].metadata["converter"](phi)

    # 2) handle negative distances
    r_pred = r < jnp.zeros_like(r)
    r = jnp.where(r_pred, -r, r)
    phi = jnp.where(r_pred, phi + _180d, phi)
    theta = jnp.where(r_pred, _180d - theta, theta)

    # 3) Handle polar angle outside of [0, 180] degrees
    theta = jnp.mod(theta, _360d)  # wrap to [0, 360) deg
    theta_pred = theta < _180d
    theta = jnp.where(theta_pred, theta, _360d - theta)
    phi = jnp.where(theta_pred, phi, phi + _180d)

    # 4) Construct. This also handles the azimuthal angle wrapping
    return cls(r=r, theta=theta, phi=phi)


# ---------------------------------------------------------


@dispatch
def vector(
    cls: type[LonLatSphericalPos],
    *,
    lon: u.AbstractQuantity,
    lat: u.AbstractQuantity,
    distance: u.AbstractQuantity,
) -> LonLatSphericalPos:
    """Construct LonLatSphericalPos, allowing for out-of-range values.

    Examples
    --------
    >>> import unxt as u
    >>> import coordinax.vecs as cxv

    Let's start with a valid input:

    >>> vec = cxv.LonLatSphericalPos.from_(lon=u.Quantity(0, "deg"),
    ...                                    lat=u.Quantity(0, "deg"),
    ...                                    distance=u.Quantity(3, "km"))
    >>> print(vec)
    <LonLatSphericalPos (lon[deg], lat[deg], distance[km])
        [0 0 3]>

    The distance can be negative, which wraps the longitude by 180 degrees and
    flips the latitude:

    >>> vec = cxv.LonLatSphericalPos.from_(lon=u.Quantity(0, "deg"),
    ...                                    lat=u.Quantity(45, "deg"),
    ...                                    distance=u.Quantity(-3, "km"))
    >>> print(vec)
    <LonLatSphericalPos (lon[deg], lat[deg], distance[km])
        [180 -45   3]>

    The latitude can be outside the [-90, 90] deg range, causing the longitude
    to be shifted by 180 degrees:

    >>> vec = cxv.LonLatSphericalPos.from_(lon=u.Quantity(0, "deg"),
    ...                                    lat=u.Quantity(-100, "deg"),
    ...                                    distance=u.Quantity(3, "km"))
    >>> print(vec)
    <LonLatSphericalPos (lon[deg], lat[deg], distance[km])
        [180 -80   3]>

    >>> vec = cxv.LonLatSphericalPos.from_(lon=u.Quantity(0, "deg"),
    ...                                    lat=u.Quantity(100, "deg"),
    ...                                    distance=u.Quantity(3, "km"))
    >>> print(vec)
    <LonLatSphericalPos (lon[deg], lat[deg], distance[km])
        [180  80   3]>

    The longitude can be outside the [0, 360) deg range. This is wrapped to the
    [0, 360) deg range (actually the base constructor does this):

    >>> vec = cxv.LonLatSphericalPos.from_(lon=u.Quantity(365, "deg"),
    ...                                    lat=u.Quantity(0, "deg"),
    ...                                    distance=u.Quantity(3, "km"))
    >>> vec.lon
    Angle(Array(5, dtype=int32, ...), unit='deg')

    """
    # 1) Convert the inputs
    fields = LonLatSphericalPos.__dataclass_fields__
    lon = fields["lon"].metadata["converter"](lon)
    lat = fields["lat"].metadata["converter"](lat)
    distance = fields["distance"].metadata["converter"](distance)

    # 2) handle negative distances
    distance_pred = distance < jnp.zeros_like(distance)
    distance = qlax.select(distance_pred, -distance, distance)
    lon = qlax.select(distance_pred, lon + _180d, lon)
    lat = qlax.select(distance_pred, -lat, lat)

    # 3) Handle latitude outside of [-90, 90] degrees
    # TODO: fix when lat < -180, lat > 180
    lat_pred = lat < -_90d
    lat = qlax.select(lat_pred, -_180d - lat, lat)
    lon = qlax.select(lat_pred, lon + _180d, lon)

    lat_pred = lat > _90d
    lat = qlax.select(lat_pred, _180d - lat, lat)
    lon = qlax.select(lat_pred, lon + _180d, lon)

    # 4) Construct. This also handles the longitude wrapping
    return cls(lon=lon, lat=lat, distance=distance)


# ---------------------------------------------------------


@dispatch
def vector(
    cls: type[MathSphericalPos],
    *,
    r: u.AbstractQuantity,
    theta: u.AbstractQuantity,
    phi: u.AbstractQuantity,
) -> MathSphericalPos:
    """Construct MathSphericalPos, allowing for out-of-range values.

    Examples
    --------
    >>> import unxt as u
    >>> import coordinax.vecs as cxv

    Let's start with a valid input:

    >>> vec = cxv.MathSphericalPos.from_(r=u.Quantity(3, "km"),
    ...                                  theta=u.Quantity(90, "deg"),
    ...                                  phi=u.Quantity(0, "deg"))
    >>> print(vec)
    <MathSphericalPos (r[km], theta[deg], phi[deg])
        [ 3 90  0]>

    The radial distance can be negative, which wraps the azimuthal angle by 180
    degrees and flips the polar angle:

    >>> vec = cxv.MathSphericalPos.from_(r=u.Quantity(-3, "km"),
    ...                                  theta=u.Quantity(100, "deg"),
    ...                                  phi=u.Quantity(45, "deg"))
    >>> print(vec)
    <MathSphericalPos (r[km], theta[deg], phi[deg])
        [  3 280 135]>

    The polar angle can be outside the [0, 180] deg range, causing the azimuthal
    angle to be shifted by 180 degrees:

    >>> vec = cxv.MathSphericalPos.from_(r=u.Quantity(3, "km"),
    ...                                  theta=u.Quantity(0, "deg"),
    ...                                  phi=u.Quantity(190, "deg"))
    >>> print(vec)
    <MathSphericalPos (r[km], theta[deg], phi[deg])
        [  3 180 170]>

    The azimuth can be outside the [0, 360) deg range. This is wrapped to the
    [0, 360) deg range (actually the base constructor does this):

    >>> vec = cxv.MathSphericalPos.from_(r=u.Quantity(3, "km"),
    ...                                  theta=u.Quantity(365, "deg"),
    ...                                  phi=u.Quantity(90, "deg"))
    >>> vec.theta
    Angle(Array(5, dtype=int32, ...), unit='deg')

    """
    # 1) Convert the inputs
    fields = MathSphericalPos.__dataclass_fields__
    r = fields["r"].metadata["converter"](r)
    theta = fields["theta"].metadata["converter"](theta)
    phi = fields["phi"].metadata["converter"](phi)

    # 2) handle negative distances
    r_pred = r < jnp.zeros_like(r)
    r = jnp.where(r_pred, -r, r)
    theta = jnp.where(r_pred, theta + _180d, theta)
    phi = jnp.where(r_pred, _180d - phi, phi)

    # 3) Handle polar angle outside of [0, 180] degrees
    phi = jnp.mod(phi, _360d)  # wrap to [0, 360) deg
    phi_pred = phi < _180d
    phi = jnp.where(phi_pred, phi, _360d - phi)
    theta = jnp.where(phi_pred, theta, theta + _180d)

    # 4) Construct. This also handles the azimuthal angle wrapping
    return cls(r=r, theta=theta, phi=phi)


###############################################################################
# Vector Transformation

# =============================================================================
# `vconvert_impl`


@dispatch
def vconvert(
    to_vector: type[AbstractPos3D],
    from_vector: type[AbstractPos3D],
    params: ct.ParamsDict,
    /,
    *,
    in_aux: ct.OptAuxDict = None,
    out_aux: ct.OptAuxDict = None,
    units: ct.OptUSys = None,
) -> tuple[ct.ParamsDict, ct.AuxDict]:
    """AbstractPos -> CartesianPos3D -> AbstractPos."""
    params, aux = vconvert(
        CartesianPos3D, from_vector, params, in_aux=in_aux, out_aux=None, units=units
    )
    params, aux = vconvert(
        to_vector, CartesianPos3D, params, in_aux=aux, out_aux=out_aux, units=units
    )
    return params, aux


@dispatch
@ft.partial(jax.jit, static_argnums=(0, 1), static_argnames=("units",))
@wrap_vconvert_impl_params
def vconvert(
    to_vector: type[CylindricalPos],
    from_vector: type[CartesianPos3D],
    p: ct.ParamsDict,
    /,
    *,
    in_aux: ct.OptAuxDict = None,
    out_aux: ct.OptAuxDict = None,
    units: ct.OptUSys = None,
) -> tuple[ct.ParamsDict, ct.AuxDict]:
    """CartesianPos3D -> CylindricalPos.

    Examples
    --------
    >>> import unxt as u
    >>> import coordinax.vecs as cxv

    >>> cart = {"x": 1, "y": 2, "z": 3}
    >>> cxv.vconvert(cxv.CylindricalPos, cxv.CartesianPos3D, cart)
    ({'phi': Array(1.1071488, dtype=float32, ...),
      'rho': Array(2.236068, dtype=float32, ...),
      'z': Array(3, dtype=int32, ...)},
     {})

    >>> cart = {"x": u.Quantity(1, "km"), "y": u.Quantity(2, "km"),
    ...         "z": u.Quantity(3, "km")}
    >>> cxv.vconvert(cxv.CylindricalPos, cxv.CartesianPos3D, cart)
    ({'phi': Quantity['angle'](Array(1.1071488, dtype=float32, ...), unit='rad'),
      'rho': Quantity['length'](Array(2.236068, dtype=float32, ...), unit='km'),
      'z': Quantity['length'](Array(3, dtype=int32, ...), unit='km')},
     {})

    """
    rho = jnp.hypot(p["x"], p["y"])
    phi = jnp.atan2(p["y"], p["x"])
    return {"rho": rho, "phi": phi, "z": p["z"]}, combine_aux(in_aux, out_aux)


@dispatch
@ft.partial(jax.jit, static_argnums=(0, 1), static_argnames=("units",))
@wrap_vconvert_impl_params
def vconvert(
    to_vector: type[SphericalPos],
    from_vector: type[CartesianPos3D],
    p: ct.ParamsDict,
    /,
    *,
    in_aux: ct.OptAuxDict = None,
    out_aux: ct.OptAuxDict = None,
    units: ct.OptUSys = None,
) -> tuple[ct.ParamsDict, ct.AuxDict]:
    """CartesianPos3D -> SphericalPos.

    Examples
    --------
    >>> import coordinax.vecs as cxv
    >>> cart = {"x": 1, "y": 2, "z": 3}
    >>> cxv.vconvert(cxv.SphericalPos, cxv.CartesianPos3D, cart)
    ({'phi': Array(1.1071488, dtype=float32, ...),
      'r': Array(3.7416575, dtype=float32, ...),
      'theta': Array(0.64052236, dtype=float32, ...)},
     {})

    """
    r = jnp.sqrt(p["x"] ** 2 + p["y"] ** 2 + p["z"] ** 2)
    theta = jnp.acos(p["z"] / r)
    phi = jnp.atan2(p["y"], p["x"])
    return {"r": r, "theta": theta, "phi": phi}, combine_aux(in_aux, out_aux)


@dispatch.multi(
    (type[LonLatSphericalPos], type[CartesianPos3D], ct.ParamsDict),
    (type[MathSphericalPos], type[CartesianPos3D], ct.ParamsDict),
)
def vconvert(
    to_vector: type[AbstractSphericalPos],
    from_vector: type[CartesianPos3D],
    p: ct.ParamsDict,
    /,
    *,
    in_aux: ct.OptAuxDict = None,
    out_aux: ct.OptAuxDict = None,
    units: ct.OptUSys = None,
) -> tuple[ct.ParamsDict, ct.AuxDict]:
    """CartesianPos3D -> AbstractSphericalPos.

    Examples
    --------
    >>> import coordinax.vecs as cxv
    >>> params = {"x": 1, "y": 2, "z": 3}
    >>> usys = u.unitsystem("km", "deg")

    >>> cxv.vconvert(cxv.LonLatSphericalPos, cxv.CartesianPos3D,
    ...                   params, units=usys)
    ({'distance': Array(3.7416575, dtype=float32, ...),
      'lat': Array(53.300774, dtype=float32, ...),
      'lon': Array(63.43495, dtype=float32, ...)},
     {})

    >>> cxv.vconvert(cxv.MathSphericalPos, cxv.CartesianPos3D,
    ...                   params, units=usys)
    ({'phi': Array(36.69923, dtype=float32, ...),
      'r': Array(3.7416575, dtype=float32, ...),
      'theta': Array(63.43495, dtype=float32, ...)},
     {})

    """
    p, aux = vconvert(
        SphericalPos, from_vector, p, in_aux=in_aux, out_aux=None, units=units
    )
    p, aux = vconvert(
        to_vector, SphericalPos, p, in_aux=aux, out_aux=out_aux, units=units
    )
    return p, aux


@dispatch
@ft.partial(jax.jit, static_argnums=(0, 1), static_argnames=("units",))
@wrap_vconvert_impl_params
def vconvert(
    to_vector: type[CartesianPos3D],
    from_vector: type[CylindricalPos],
    p: ct.ParamsDict,
    /,
    *,
    in_aux: ct.OptAuxDict = None,
    out_aux: ct.OptAuxDict = None,
    units: ct.OptUSys = None,
) -> tuple[ct.ParamsDict, ct.AuxDict]:
    """CylindricalPos -> CartesianPos3D.

    Examples
    --------
    >>> import coordinax.vecs as cxv
    >>> cyl = {"rho": 1, "phi": 90, "z": 1}
    >>> usys = u.unitsystem("km", "deg")

    >>> cxv.vconvert(cxv.CartesianPos3D, cxv.CylindricalPos, cyl, units=usys)
    ({'x': Array(-4.371139e-08, dtype=float32, ...),
      'y': Array(1., dtype=float32, ...),
      'z': Array(1, dtype=int32, ...)},
     {})

    """
    x = p["rho"] * jnp.cos(p["phi"])
    y = p["rho"] * jnp.sin(p["phi"])
    return {"x": x, "y": y, "z": p["z"]}, combine_aux(in_aux, out_aux)


@dispatch
@ft.partial(jax.jit, static_argnums=(0, 1), static_argnames=("units",))
@wrap_vconvert_impl_params
def vconvert(
    to_vector: type[SphericalPos],
    from_vector: type[CylindricalPos],
    p: ct.ParamsDict,
    /,
    *,
    in_aux: ct.OptAuxDict = None,
    out_aux: ct.OptAuxDict = None,
    units: ct.OptUSys = None,
) -> tuple[ct.ParamsDict, ct.AuxDict]:
    """CylindricalPos -> SphericalPos.

    Examples
    --------
    >>> import coordinax.vecs as cxv
    >>> cyl = {"rho": 1, "phi": 90, "z": 1}
    >>> usys = u.unitsystem("km", "deg")

    >>> cxv.vconvert(cxv.SphericalPos, cxv.CylindricalPos, cyl, units=usys)
    ({'phi': Array(90., dtype=float32, ...),
      'r': Array(1.4142135, dtype=float32, ...),
      'theta': Array(45..., dtype=float32, ...)},
     {})

    """
    r = jnp.hypot(p["rho"], p["z"])
    theta = jnp.acos(p["z"] / r)
    return {"r": r, "theta": theta, "phi": p["phi"]}, combine_aux(in_aux, out_aux)


@dispatch.multi(
    (type[LonLatSphericalPos], type[CylindricalPos], ct.ParamsDict),
    (type[MathSphericalPos], type[CylindricalPos], ct.ParamsDict),
)
@ft.partial(jax.jit, static_argnums=(0, 1), static_argnames=("units",))
def vconvert(
    to_vector: type[AbstractSphericalPos],
    from_vector: type[CylindricalPos],
    p: ct.ParamsDict,
    /,
    *,
    in_aux: ct.OptAuxDict = None,
    out_aux: ct.OptAuxDict = None,
    units: ct.OptUSys = None,
) -> tuple[ct.ParamsDict, ct.AuxDict]:
    """CylindricalPos -> AbstractSphericalPos.

    Examples
    --------
    >>> import coordinax.vecs as cxv
    >>> cyl = {"rho": 1, "phi": 90, "z": 1}
    >>> usys = u.unitsystem("km", "deg")

    >>> cxv.vconvert(cxv.LonLatSphericalPos, cxv.CylindricalPos, cyl, units=usys)
    ({'distance': Array(1.4142135, dtype=float32, ...),
      'lat': Array(45., dtype=float32, ...),
      'lon': Array(90., dtype=float32, ...)},
     {})

    >>> cxv.vconvert(cxv.MathSphericalPos, cxv.CylindricalPos, cyl, units=usys)
    ({'phi': Array(45..., dtype=float32, ...),
      'r': Array(1.4142135, dtype=float32, ...),
      'theta': Array(90., dtype=float32, ...)},
     {})

    """
    params, aux = vconvert(
        SphericalPos, from_vector, p, in_aux=in_aux, out_aux=out_aux, units=units
    )
    params, aux = vconvert(
        to_vector, SphericalPos, params, in_aux=aux, out_aux=out_aux, units=units
    )
    return params, aux


@dispatch
@ft.partial(jax.jit, static_argnums=(0, 1), static_argnames=("units",))
@wrap_vconvert_impl_params
def vconvert(
    to_vector: type[CartesianPos3D],
    from_vector: type[SphericalPos],
    p: ct.ParamsDict,
    /,
    *,
    in_aux: ct.OptAuxDict = None,
    out_aux: ct.OptAuxDict = None,
    units: ct.OptUSys = None,
) -> tuple[ct.ParamsDict, ct.AuxDict]:
    """CylindricalPos -> CartesianPos3D.

    Examples
    --------
    >>> import coordinax.vecs as cxv
    >>> sph = {"r": 1, "theta": 90, "phi": 90}
    >>> usys = u.unitsystem("km", "deg")

    >>> cxv.vconvert(cxv.CartesianPos3D, cxv.SphericalPos, sph, units=usys)
    ({'x': Array(-4.371139e-08, dtype=float32, ...),
      'y': Array(1., dtype=float32, ...),
      'z': Array(-4.371139e-08, dtype=float32, ...)},
     {})

    """
    x = p["r"] * jnp.sin(p["theta"]) * jnp.cos(p["phi"])
    y = p["r"] * jnp.sin(p["theta"]) * jnp.sin(p["phi"])
    z = p["r"] * jnp.cos(p["theta"])
    return {"x": x, "y": y, "z": z}, combine_aux(in_aux, out_aux)


@dispatch
@ft.partial(jax.jit, static_argnums=(0, 1), static_argnames=("units",))
@wrap_vconvert_impl_params
def vconvert(
    to_vector: type[CylindricalPos],
    from_vector: type[SphericalPos],
    p: ct.ParamsDict,
    /,
    *,
    in_aux: ct.OptAuxDict = None,
    out_aux: ct.OptAuxDict = None,
    units: ct.OptUSys = None,
) -> tuple[ct.ParamsDict, ct.AuxDict]:
    """SphericalPos -> CylindricalPos.

    Examples
    --------
    >>> import coordinax.vecs as cxv
    >>> sph = {"r": 1, "theta": 90, "phi": 90}
    >>> usys = u.unitsystem("km", "deg")

    >>> cxv.vconvert(cxv.CylindricalPos, cxv.SphericalPos, sph, units=usys)
    ({'phi': Array(90., dtype=float32, ...),
      'rho': Array(1., dtype=float32, ...),
      'z': Array(-4.371139e-08, dtype=float32, ...)},
     {})

    """
    rho = jnp.abs(p["r"]) * jnp.sin(p["theta"])
    phi = p["phi"]
    z = p["r"] * jnp.cos(p["theta"])
    return {"rho": rho, "phi": phi, "z": z}, combine_aux(in_aux, out_aux)


@dispatch
@ft.partial(jax.jit, static_argnums=(0, 1), static_argnames=("units",))
@wrap_vconvert_impl_params
def vconvert(
    to_vector: type[LonLatSphericalPos],
    from_vector: type[SphericalPos],
    p: ct.ParamsDict,
    /,
    *,
    in_aux: ct.OptAuxDict = None,
    out_aux: ct.OptAuxDict = None,
    units: ct.OptUSys = None,
) -> tuple[ct.ParamsDict, ct.AuxDict]:
    """SphericalPos -> LonLatSphericalPos.

    Examples
    --------
    >>> import coordinax.vecs as cxv
    >>> sph = {"r": 1, "theta": 90, "phi": 90}
    >>> usys = u.unitsystem("km", "deg")

    >>> cxv.vconvert(cxv.LonLatSphericalPos, cxv.SphericalPos, sph, units=usys)
    ({'distance': Array(1, dtype=int32, ...),
      'lat': Array(3.2016512e-06, dtype=float32, ...),
      'lon': Array(90., dtype=float32, ...)},
     {})

    """
    if isinstance(p["theta"], u.AbstractQuantity):
        lat = u.Quantity(90, "deg") - p["theta"]
    else:
        lat = jnp.pi / 2 - p["theta"]

    aux = combine_aux(in_aux, out_aux)
    return {"lon": p["phi"], "lat": lat, "distance": p["r"]}, aux


@dispatch.multi(
    (type[LonLatSphericalVel], type[SphericalVel], ct.ParamsDict),
    (type[LonLatSphericalAcc], type[SphericalAcc], ct.ParamsDict),
)
@wrap_vconvert_impl_params
def vconvert(
    to_vector: type[AbstractVector],
    from_vector: type[AbstractVector],
    p: ct.ParamsDict,
    /,
    *,
    in_aux: ct.OptAuxDict = None,
    out_aux: ct.OptAuxDict = None,
    units: ct.OptUSys = None,
) -> tuple[ct.ParamsDict, ct.AuxDict]:
    """SphericalVel/Acc -> LonLatSphericalVel/Acc.

    Examples
    --------
    >>> import unxt as u
    >>> import coordinax.vecs as cxv

    >>> p = {"r": u.Quantity(1, "km/s"),
    ...      "theta": u.Quantity(10, "deg/s"),
    ...      "phi": u.Quantity(20, "deg/s")}

    >>> cxv.vconvert(cxv.LonLatSphericalVel, cxv.SphericalVel, p)
    ({'distance': Quantity[...](Array(1, dtype=int32, ...), unit='km / s'),
      'lat': Quantity[...](Array(-10, dtype=int32, ...), unit='deg / s'),
      'lon': Quantity[...](Array(20, dtype=int32, ...), unit='deg / s')},
     {})

    >>> x = cxv.SphericalVel(**p)
    >>> y = cxv.vconvert(cxv.LonLatSphericalVel, x)
    >>> print(y)
    <LonLatSphericalVel (lon[deg / s], lat[deg / s], distance[km / s])
        [ 20 -10   1]>

    >>> p = {"r": u.Quantity(1, "km/s2"),
    ...      "theta": u.Quantity(10, "deg/s2"),
    ...      "phi": u.Quantity(20, "deg/s2")}

    >>> cxv.vconvert(cxv.LonLatSphericalAcc, cxv.SphericalAcc, p)
    ({'distance': Quantity[...](Array(1, dtype=int32, ...), unit='km / s2'),
      'lat': Quantity[...](Array(-10, dtype=int32, ...), unit='deg / s2'),
      'lon': Quantity[...](Array(20, dtype=int32, ...), unit='deg / s2')},
     {})

    >>> x = cxv.SphericalAcc(**p)
    >>> y = cxv.vconvert(cxv.LonLatSphericalAcc, x)
    >>> print(y)
    <LonLatSphericalAcc (lon[deg / s2], lat[deg / s2], distance[km / s2])
        [ 20 -10   1]>

    """
    newp = {"distance": p["r"], "lat": -p["theta"], "lon": p["phi"]}
    return newp, (out_aux or {})


@dispatch.multi(
    (type[MathSphericalPos], type[SphericalPos], ct.ParamsDict),
    (type[SphericalPos], type[MathSphericalPos], ct.ParamsDict),
    (type[MathSphericalVel], type[SphericalVel], ct.ParamsDict),
    (type[SphericalVel], type[MathSphericalVel], ct.ParamsDict),
    (type[MathSphericalAcc], type[SphericalAcc], ct.ParamsDict),
    (type[SphericalAcc], type[MathSphericalAcc], ct.ParamsDict),
)
@ft.partial(jax.jit, static_argnums=(0, 1), static_argnames=("units",))
@wrap_vconvert_impl_params
def vconvert(
    to_vector: type[AbstractVector],
    from_vector: type[AbstractVector],
    p: ct.ParamsDict,
    /,
    *,
    in_aux: ct.OptAuxDict = None,
    out_aux: ct.OptAuxDict = None,
    units: ct.OptUSys = None,
) -> tuple[ct.ParamsDict, ct.AuxDict]:
    """SphericalPos <-> MathSphericalPos.

    Examples
    --------
    >>> import unxt as u
    >>> import coordinax.vecs as cxv

    Position:

    >>> p = {"r": 1, "theta": 90, "phi": 90}
    >>> usys = u.unitsystem("km", "deg")

    >>> cxv.vconvert(cxv.MathSphericalPos, cxv.SphericalPos, p, units=usys)
    ({'phi': Array(90., dtype=float32, ...),
      'r': Array(1, dtype=int32, ...),
      'theta': Array(90., dtype=float32, ...)},
     {})

    >>> cxv.vconvert(cxv.SphericalPos, cxv.MathSphericalPos, p, units=usys)
    ({'phi': Array(90., dtype=float32, ...),
      'r': Array(1, dtype=int32, ...),
      'theta': Array(90., dtype=float32, ...)},
     {})

    Velocity:

    >>> p = {"r": u.Quantity(1, "km/s"),
    ...      "theta": u.Quantity(10, "deg/s"),
    ...      "phi": u.Quantity(20, "deg/s")}

    >>> p, aux = cxv.vconvert(cxv.MathSphericalVel, cxv.SphericalVel, p)
    >>> p, aux
    ({'phi': Quantity[...](Array(10, dtype=int32, ...), unit='deg / s'),
      'r': Quantity['speed'](Array(1, dtype=int32, ...), unit='km / s'),
      'theta': Quantity[...](Array(20, dtype=int32, ...), unit='deg / s')}, {})

    >>> cxv.vconvert(cxv.SphericalVel, cxv.MathSphericalVel, p)
    ({'phi': Quantity[...](Array(20, dtype=int32, ...), unit='deg / s'),
      'r': Quantity['speed'](Array(1, dtype=int32, ...), unit='km / s'),
      'theta': Quantity[...](Array(10, dtype=int32, ...), unit='deg / s')},
     {})

    >>> x = cxv.SphericalVel(r=u.Quantity(1, "km/s"),
    ...                      theta=u.Quantity(10, "deg/s"),
    ...                      phi=u.Quantity(20, "deg/s"))
    >>> y = cxv.vconvert(cxv.MathSphericalVel, x)
    >>> print(y)
    <MathSphericalVel (r[km / s], theta[deg / s], phi[deg / s])
        [ 1 20 10]>

    >>> x = cxv.vconvert(cxv.SphericalVel, y)
    >>> print(x)
    <SphericalVel (r[km / s], theta[deg / s], phi[deg / s])
        [ 1 10 20]>

    Acceleration:

    >>> p = {"r": u.Quantity(1, "km/s2"),
    ...      "theta": u.Quantity(10, "deg/s2"),
    ...      "phi": u.Quantity(20, "deg/s2")}

    >>> p, aux = cxv.vconvert(cxv.MathSphericalAcc, cxv.SphericalAcc, p)
    >>> p, aux
    ({'phi': Quantity[...](Array(10, dtype=int32, ...), unit='deg / s2'),
      'r': Quantity[...](Array(1, dtype=int32, ...), unit='km / s2'),
      'theta': Quantity[...](Array(20, dtype=int32, ...), unit='deg / s2')},
     {})

    >>> cxv.vconvert(cxv.SphericalAcc, cxv.MathSphericalAcc, p)
    ({'phi': Quantity[...](Array(20, dtype=int32, ...), unit='deg / s2'),
      'r': Quantity[...](Array(1, dtype=int32, ...), unit='km / s2'),
      'theta': Quantity[...](Array(10, dtype=int32, ...), unit='deg / s2')},
     {})

    >>> x = cxv.SphericalAcc(r=u.Quantity(1, "km/s2"),
    ...                      theta=u.Quantity(10, "deg/s2"),
    ...                      phi=u.Quantity(20, "deg/s2"))
    >>> y = cxv.vconvert(cxv.MathSphericalAcc, x)
    >>> print(y)
    <MathSphericalAcc (r[km / s2], theta[deg / s2], phi[deg / s2])
        [ 1 20 10]>

    >>> x = cxv.vconvert(cxv.SphericalAcc, y)
    >>> print(x)
    <SphericalAcc (r[km / s2], theta[deg / s2], phi[deg / s2])
        [ 1 10 20]>

    """
    outp = {"r": p["r"], "theta": p["phi"], "phi": p["theta"]}
    return outp, combine_aux(in_aux, out_aux)


@dispatch
@ft.partial(jax.jit, static_argnums=(0, 1), static_argnames=("units",))
def vconvert(
    to_vector: type[CartesianPos3D],
    from_vector: type[LonLatSphericalPos],
    p: ct.ParamsDict,
    /,
    *,
    in_aux: ct.OptAuxDict = None,
    out_aux: ct.OptAuxDict = None,
    units: ct.OptUSys = None,
) -> tuple[ct.ParamsDict, ct.AuxDict]:
    """LonLatSphericalPos -> CartesianPos3D.

    Examples
    --------
    >>> import coordinax.vecs as cxv
    >>> vec = {"lon": 90, "lat": 0, "distance": 1}
    >>> usys = u.unitsystem("km", "deg")

    >>> cxv.vconvert(cxv.CartesianPos3D, cxv.LonLatSphericalPos, vec, units=usys)
    ({'x': Array(-4.371139e-08, dtype=float32, ...),
      'y': Array(1., dtype=float32, ...),
      'z': Array(-4.371139e-08, dtype=float32, ...)},
     {})

    """
    p, aux = vconvert(
        SphericalPos, from_vector, p, in_aux=in_aux, out_aux=out_aux, units=units
    )
    p, aux = vconvert(
        to_vector, SphericalPos, p, in_aux=aux, out_aux=out_aux, units=units
    )
    return p, aux


@dispatch
@ft.partial(jax.jit, static_argnums=(0, 1), static_argnames=("units",))
def vconvert(
    to_vector: type[CylindricalPos],
    from_vector: type[LonLatSphericalPos],
    p: ct.ParamsDict,
    /,
    *,
    in_aux: ct.OptAuxDict = None,
    out_aux: ct.OptAuxDict = None,
    units: ct.OptUSys = None,
) -> tuple[ct.ParamsDict, ct.AuxDict]:
    """LonLatSphericalPos -> CylindricalPos.

    Examples
    --------
    >>> import coordinax.vecs as cxv
    >>> sph = {"lon": 90, "lat": 0, "distance": 1}
    >>> usys = u.unitsystem("km", "deg")

    >>> cxv.vconvert(cxv.CylindricalPos, cxv.LonLatSphericalPos, sph, units=usys)
    ({'phi': Array(90., dtype=float32, ...),
      'rho': Array(1., dtype=float32, ...),
      'z': Array(-4.371139e-08, dtype=float32, ...)},
     {})

    """
    p, aux = vconvert(
        SphericalPos, from_vector, p, in_aux=in_aux, out_aux=None, units=units
    )
    p, aux = vconvert(
        to_vector, SphericalPos, p, in_aux=aux, out_aux=out_aux, units=units
    )
    return p, aux


@dispatch
@ft.partial(jax.jit, static_argnums=(0, 1), static_argnames=("units",))
@wrap_vconvert_impl_params
def vconvert(
    to_vector: type[SphericalPos],
    from_vector: type[LonLatSphericalPos],
    p: ct.ParamsDict,
    /,
    *,
    in_aux: ct.OptAuxDict = None,
    out_aux: ct.OptAuxDict = None,
    units: ct.OptUSys = None,
) -> tuple[ct.ParamsDict, ct.AuxDict]:
    """LonLatSphericalPos -> SphericalPos.

    Examples
    --------
    >>> import coordinax.vecs as cxv
    >>> vec = {"lon": 90, "lat": 0, "distance": 1}
    >>> usys = u.unitsystem("km", "deg")

    >>> cxv.vconvert(cxv.SphericalPos, cxv.LonLatSphericalPos, vec, units=usys)
    ({'phi': Array(90., dtype=float32, ...,
      'r': Array(1, dtype=int32, ...,
      'theta': Array(90., dtype=float32, ...},
     {})

    """
    if isinstance(p["lat"], u.AbstractQuantity):
        theta = u.Quantity(90, "deg") - p["lat"]
    else:
        theta = jnp.pi / 2 - p["lat"]

    outp = {"r": p["distance"], "theta": theta, "phi": p["lon"]}
    return outp, combine_aux(in_aux, out_aux)


@dispatch.multi(
    (type[SphericalVel], type[LonLatSphericalVel], ct.ParamsDict),
    (type[SphericalAcc], type[LonLatSphericalAcc], ct.ParamsDict),
)
@wrap_vconvert_impl_params
def vconvert(
    to_vector: type[AbstractVector],
    from_vector: type[AbstractVector],
    p: ct.ParamsDict,
    /,
    *,
    in_aux: ct.OptAuxDict = None,
    out_aux: ct.OptAuxDict = None,
    units: ct.OptUSys = None,
) -> tuple[ct.ParamsDict, ct.AuxDict]:
    """LonLatSphericalVel -> SphericalVel.

    Examples
    --------
    >>> import unxt as u
    >>> import coordinax.vecs as cxv

    >>> p = {"lon": u.Quantity(90, "deg/s"),
    ...      "lat": u.Quantity(0, "deg/s"),
    ...      "distance": u.Quantity(1, "km/s")}
    >>> cxv.vconvert(cxv.SphericalVel, cxv.LonLatSphericalVel, p)
    ({'r': Quantity[...](Array(1, dtype=int32, ...), unit='km / s'),
      'theta': Quantity[...](Array(0, dtype=int32, ...), unit='deg / s'),
      'phi': Quantity[...](Array(90, dtype=int32, ...), unit='deg / s')},
     {})

    >>> x = cxv.LonLatSphericalVel(**p)
    >>> y = cxv.vconvert(cxv.SphericalVel, x)
    >>> print(y)
    <SphericalVel (r[km / s], theta[deg / s], phi[deg / s])
        [ 1  0 90]>

    >>> p = {"lon": u.Quantity(90, "deg/s2"),
    ...      "lat": u.Quantity(0, "deg/s2"),
    ...      "distance": u.Quantity(1, "km/s2")}
    >>> cxv.vconvert(cxv.SphericalAcc, cxv.LonLatSphericalAcc, p)
    ({'r': Quantity[...](Array(1, dtype=int32, ...), unit='km / s2'),
      'theta': Quantity[...](Array(0, dtype=int32, ...), unit='deg / s2'),
      'phi': Quantity[...](Array(90, dtype=int32, ...), unit='deg / s2')},
     {})

    >>> x = cxv.LonLatSphericalAcc(**p)
    >>> y = cxv.vconvert(cxv.SphericalAcc, x)
    >>> print(y)
    <SphericalAcc (r[km / s2], theta[deg / s2], phi[deg / s2])
        [ 1  0 90]>

    """
    newp = {"r": p["distance"], "theta": -p["lat"], "phi": p["lon"]}
    return newp, (out_aux or {})


@dispatch
@ft.partial(jax.jit, static_argnums=(0, 1), static_argnames=("units",))
@wrap_vconvert_impl_params
def vconvert(
    to_vector: type[CartesianPos3D],
    from_vector: type[MathSphericalPos],
    p: ct.ParamsDict,
    /,
    *,
    in_aux: ct.OptAuxDict = None,
    out_aux: ct.OptAuxDict = None,
    units: ct.OptUSys = None,
) -> tuple[ct.ParamsDict, ct.AuxDict]:
    """MathSphericalPos -> CartesianPos3D.

    Examples
    --------
    >>> import coordinax.vecs as cxv
    >>> vec = {"r": 1, "theta": 90, "phi": 90}
    >>> usys = u.unitsystem("km", "deg")

    >>> cxv.vconvert(cxv.CartesianPos3D, cxv.MathSphericalPos, vec, units=usys)
    ({'x': Array(-4.371139e-08, dtype=float32, ...),
      'y': Array(1., dtype=float32, ...),
      'z': Array(-4.371139e-08, dtype=float32, ...)},
     {})

    """
    x = p["r"] * jnp.sin(p["phi"]) * jnp.cos(p["theta"])
    y = p["r"] * jnp.sin(p["phi"]) * jnp.sin(p["theta"])
    z = p["r"] * jnp.cos(p["phi"])
    return {"x": x, "y": y, "z": z}, combine_aux(in_aux, out_aux)


@dispatch
@ft.partial(jax.jit, static_argnums=(0, 1), static_argnames=("units",))
@wrap_vconvert_impl_params
def vconvert(
    to_vector: type[CylindricalPos],
    from_vector: type[MathSphericalPos],
    p: ct.ParamsDict,
    /,
    *,
    in_aux: ct.OptAuxDict = None,
    out_aux: ct.OptAuxDict = None,
    units: ct.OptUSys = None,
) -> tuple[ct.ParamsDict, ct.AuxDict]:
    """MathSphericalPos -> CylindricalPos.

    Examples
    --------
    >>> import coordinax.vecs as cxv
    >>> vec = {"r": 1, "theta": 90, "phi": 90}
    >>> usys = u.unitsystem("km", "deg")

    >>> cxv.vconvert(cxv.CylindricalPos, cxv.MathSphericalPos, vec, units=usys)
    ({'phi': Array(90., dtype=float32, ...),
      'rho': Array(1., dtype=float32, ...),
      'z': Array(-4.371139e-08, dtype=float32, ...)},
     {})

    """
    rho = jnp.abs(p["r"]) * jnp.sin(p["phi"])
    z = p["r"] * jnp.cos(p["phi"])
    return {"rho": rho, "phi": p["theta"], "z": z}, combine_aux(in_aux, out_aux)


@dispatch
@ft.partial(jax.jit, static_argnums=(0, 1), static_argnames=("units",))
@wrap_vconvert_impl_params
def vconvert(
    to_vector: type[CylindricalPos],
    from_vector: type[ProlateSpheroidalPos],
    p: ct.ParamsDict,
    /,
    *,
    in_aux: ct.OptAuxDict = None,
    out_aux: ct.OptAuxDict = None,
    units: ct.OptUSys = None,
) -> tuple[ct.ParamsDict, ct.AuxDict]:
    """ProlateSpheroidalPos -> CylindricalPos.

    Examples
    --------
    >>> import unxt as u
    >>> import coordinax.vecs as cxv
    >>> vec = {"mu": 1, "nu": 0.2, "phi": 90}
    >>> in_aux = {"Delta": 0.5}
    >>> usys = u.unitsystem("km", "deg")

    >>> cxv.vconvert(cxv.CylindricalPos, cxv.ProlateSpheroidalPos,
    ...                   vec, in_aux=in_aux, units=usys)
    ({'phi': Array(90., dtype=float32, ...),
      'rho': Array(0.38729832, dtype=float32, ...),
      'z': Array(0.8944272, dtype=float32, ...)},
     {})

    TODO: example with Delta as a Quantity

    """
    Delta2 = in_aux["Delta"] ** 2
    nu_D2 = jnp.abs(p["nu"]) / Delta2
    rho = jnp.sqrt((p["mu"] - Delta2) * (1 - nu_D2))
    z = jnp.sqrt(p["mu"] * nu_D2) * jnp.sign(p["nu"])
    return {"rho": rho, "phi": p["phi"], "z": z}, {}


@dispatch
@ft.partial(jax.jit, static_argnums=(0, 1), static_argnames=("units",))
@wrap_vconvert_impl_params
def vconvert(
    to_vector: type[ProlateSpheroidalPos],
    from_vector: type[CylindricalPos],
    p: ct.ParamsDict,
    /,
    *,
    in_aux: ct.OptAuxDict = None,
    out_aux: ct.OptAuxDict = None,
    units: ct.OptUSys = None,
) -> tuple[ct.ParamsDict, ct.AuxDict]:
    """CylindricalPos -> ProlateSpheroidalPos.

    Examples
    --------
    >>> import unxt as u
    >>> import coordinax.vecs as cxv

    >>> vec = {"rho": 1, "phi": 90, "z": 1}
    >>> out_aux = {"Delta": 0.5}
    >>> usys = u.unitsystem("km", "deg")

    >>> cxv.vconvert(cxv.ProlateSpheroidalPos, cxv.CylindricalPos,
    ...                   vec, out_aux=out_aux, units=usys)
    ({'mu': Array(2.1327822, dtype=float32, ...),
      'nu': Array(0.11721778, dtype=float32, ...),
      'phi': Array(90., dtype=float32, ...)},
     {'Delta': Array(0.5, dtype=float32, ...)})

    # <ProlateSpheroidalPos (mu[km2], nu[km2], phi[deg])
    #     [ 2.133  0.117 90.   ]>

    TODO: example with Delta as a Quantity

    """
    Delta = eqx.error_if(
        out_aux.get("Delta"),
        "Delta" not in out_aux,
        "Delta must be provided for ProlateSpheroidalPos.",
    )
    R2 = p["rho"] ** 2
    z2 = p["z"] ** 2
    Delta2 = Delta**2

    sum_ = R2 + z2 + Delta2
    diff_ = R2 + z2 - Delta2

    # compute D = sqrt((R² + z² - Δ²)² + 4R²Δ²)
    D = jnp.sqrt(diff_**2 + 4 * R2 * Delta2)

    # handle special cases for R=0 or z=0
    D = jnp.where(p["z"] == 0, sum_, D)
    D = jnp.where(p["rho"] == 0, jnp.abs(diff_), D)

    # compute mu and nu depending on sign of diff_ - avoids dividing by a small number
    pos_mu_minus_delta = 0.5 * (D + diff_)
    pos_delta_minus_nu = Delta2 * R2 / pos_mu_minus_delta

    neg_delta_minus_nu = 0.5 * (D - diff_)
    neg_mu_minus_delta = Delta2 * R2 / neg_delta_minus_nu

    # Select based on condition
    mu_minus_delta = jnp.where(diff_ >= 0, pos_mu_minus_delta, neg_mu_minus_delta)
    delta_minus_nu = jnp.where(diff_ >= 0, pos_delta_minus_nu, neg_delta_minus_nu)

    # compute mu and nu:
    mu = Delta2 + mu_minus_delta
    abs_nu = 2 * Delta2 / (sum_ + D) * z2

    # for numerical stability when Delta^2-|nu| is small
    abs_nu = jnp.where(abs_nu * 2 > Delta2, Delta2 - delta_minus_nu, abs_nu)

    nu = abs_nu * jnp.sign(p["z"])

    return {"mu": mu, "nu": nu, "phi": p["phi"]}, out_aux


@dispatch
@ft.partial(jax.jit, static_argnums=(0, 1), static_argnames=("units",))
def vconvert(
    to_vector: type[ProlateSpheroidalPos],
    from_vector: type[ProlateSpheroidalPos],
    p: ct.ParamsDict,
    /,
    *,
    in_aux: ct.AuxDict,
    out_aux: ct.OptAuxDict = None,
    units: ct.OptUSys = None,
) -> tuple[ct.ParamsDict, ct.AuxDict]:
    """ProlateSpheroidalPos -> ProlateSpheroidalPos.

    Examples
    --------
    >>> import coordinax.vecs as cxv
    >>> vec = {"mu": 1, "nu": 0.2, "phi": 90}
    >>> in_aux = {"Delta": 0.5}
    >>> usys = u.unitsystem("km", "deg")

    Self-transforms can change the focal length:

    >>> out_aux = {"Delta": 0.8}
    >>> cxv.vconvert(cxv.ProlateSpheroidalPos, cxv.ProlateSpheroidalPos,
    ...                   vec, in_aux=in_aux, out_aux=out_aux, units=usys)
    ({'mu': Array(1.1414464, dtype=float32, ...),
      'nu': Array(0.44855377, dtype=float32, ...),
      'phi': Array(90., dtype=float32, ...)},
     {'Delta': Array(0.8, dtype=float32, ...)})

    Without changing the focal length, no transform is done:

    >>> cxv.vconvert(cxv.ProlateSpheroidalPos, cxv.ProlateSpheroidalPos,
    ...                   vec, in_aux=in_aux, units=usys)
    ({'mu': Array(1, dtype=int32, ...),
      'nu': Array(0.2, dtype=float32, ...),
      'phi': Array(90, dtype=int32, ...)},
     {'Delta': Array(0.5, dtype=float32, ...)})

    """
    out_aux = out_aux or {}
    if out_aux.get("Delta", None) is None:
        return p, combine_aux(in_aux, out_aux)

    # If Delta is provided, we can proceed with the conversion
    p, aux = vconvert(
        CylindricalPos, from_vector, p, in_aux=in_aux, out_aux=None, units=units
    )
    p, aux = vconvert(
        to_vector, CylindricalPos, p, in_aux=aux, out_aux=out_aux, units=units
    )
    return p, aux


@dispatch
@ft.partial(jax.jit, static_argnums=(0, 1), static_argnames=("units",))
def vconvert(
    to_vector: type[ProlateSpheroidalPos],
    from_vector: type[AbstractPos3D,],
    p: ct.ParamsDict,
    /,
    *,
    in_aux: ct.OptAuxDict = None,
    out_aux: ct.AuxDict,
    units: ct.OptUSys = None,
) -> tuple[ct.ParamsDict, ct.AuxDict]:
    """AbstractPos3D -> ProlateSpheroidalPos.

    Examples
    --------
    >>> import unxt as u
    >>> import coordinax.vecs as cxv

    >>> vec = {"x": 1, "y": 2, "z": 3}
    >>> out_aux = {"Delta": 0.5}
    >>> usys = u.unitsystem("km", "deg")

    >>> cxv.vconvert(cxv.ProlateSpheroidalPos, cxv.CartesianPos3D,
    ...                   vec, out_aux=out_aux, units=usys)
    ({'mu': Array(14.090316, dtype=float32, ...),
      'nu': Array(0.15968415, dtype=float32, ...),
      'phi': Array(63.43495, dtype=float32, ...)},
     {'Delta': Array(0.5, dtype=float32, ...)})

    """
    p, aux = vconvert(
        CylindricalPos, from_vector, p, in_aux=in_aux, out_aux=None, units=units
    )
    p, aux = vconvert(
        to_vector, CylindricalPos, p, in_aux=aux, out_aux=out_aux, units=units
    )
    return p, aux


# =============================================================================
# LonLatSphericalVel


@dispatch
def vconvert(
    target: type[LonCosLatSphericalVel],
    current: AbstractVel3D,
    position: AbstractPos | u.AbstractQuantity,
    /,
    **kwargs: Any,
) -> LonCosLatSphericalVel:
    """AbstractVel3D -> LonCosLatSphericalVel.

    Examples
    --------
    >>> import quaxed.numpy as jnp
    >>> import unxt as u
    >>> import coordinax.vecs as cxv

    >>> q = cxv.LonLatSphericalPos(lon=u.Quantity(15, "deg"),
    ...                            lat=u.Quantity(10, "deg"),
    ...                            distance=u.Quantity(1.5, "km"))
    >>> p = cxv.LonLatSphericalVel(lon=u.Quantity(7, "mas/yr"),
    ...                            lat=u.Quantity(0, "deg/Gyr"),
    ...                            distance=u.Quantity(-5, "km/s"))
    >>> newp = cxv.vconvert(cxv.LonCosLatSphericalVel, p, q)
    >>> print(newp)
    <LonCosLatSphericalVel (lon_coslat[mas / yr], lat[deg / Gyr], distance[km / s])
        [ 6.894  0.    -5.   ]>

    """
    # Parse the position to an AbstractPos
    posvec = (
        position
        if isinstance(position, AbstractPos)
        else current.time_antiderivative_cls.cartesian_type.from_(position)
    )

    # Transform the differential to LonLatSphericalVel
    current = vconvert(LonLatSphericalVel, current, posvec)

    # Transform the position to the required type
    posvec = vconvert(current.time_antiderivative_cls, posvec)

    # Calculate the differential in the new system
    return target(
        lon_coslat=current.lon * jnp.cos(posvec.lat),
        lat=current.lat,
        distance=current.distance,
    )


@dispatch
def vconvert(
    target: type[LonLatSphericalVel],
    current: LonCosLatSphericalVel,
    position: AbstractPos | u.Quantity["length"],
    /,
    **kwargs: Any,
) -> LonLatSphericalVel:
    """LonCosLatSphericalVel -> LonLatSphericalVel."""
    # Parse the position to an AbstractPos
    posvec = (
        position
        if isinstance(position, AbstractPos)
        else current.time_antiderivative_cls.cartesian_type.from_(position)
    )

    # Transform the position to the required type
    posvec = vconvert(current.time_antiderivative_cls, posvec)

    # Calculate the differential in the new system
    return target(
        lon=current.lon_coslat / jnp.cos(posvec.lat),
        lat=current.lat,
        distance=current.distance,
    )


@dispatch
def vconvert(
    target: type[AbstractVel3D],
    current: LonCosLatSphericalVel,
    position: AbstractPos | u.Quantity["length"],
    /,
    **kwargs: Any,
) -> AbstractVel3D:
    """LonCosLatSphericalVel -> AbstractVel3D."""
    # Parse the position to an AbstractPos
    posvec = (
        position
        if isinstance(position, AbstractPos)
        else current.time_antiderivative_cls.cartesian_type.from_(position)
    )
    # Transform the differential to LonLatSphericalVel
    current = vconvert(LonLatSphericalVel, current, posvec)
    # Transform the position to the required type
    return vconvert(target, current, posvec)


#####################################################################


# from coordinax.vectors.funcs
@dispatch
@ft.partial(eqx.filter_jit, inline=True)
def normalize_vector(obj: CartesianPos3D, /) -> Cartesian3D:
    """Return the norm of the vector.

    This has length 1.

    .. note::

        The unit vector is dimensionless, even if the input vector has units.
        This is because the unit vector is a ratio of two quantities: each
        component and the norm of the vector.

    Returns
    -------
    Cartesian3D
        The norm of the vector.

    Examples
    --------
    >>> import coordinax.vecs as cxv
    >>> q = cxv.CartesianPos3D.from_([1, 2, 3], "km")
    >>> print(cxv.normalize_vector(q))
    <Cartesian3D (x[], y[], z[])
        [0.267 0.535 0.802]>

    """
    norm: u.AbstractQuantity = obj.norm()  # type: ignore[misc]
    return Cartesian3D(x=obj.x / norm, y=obj.y / norm, z=obj.z / norm)


###############################################################################
# Corresponding Cartesian classes


@dispatch
def cartesian_vector_type(
    obj: type[AbstractPos3D] | AbstractPos3D, /
) -> type[CartesianPos3D]:
    """Return the corresponding Cartesian class."""
    return CartesianPos3D


@dispatch
def cartesian_vector_type(
    obj: type[AbstractVel3D] | AbstractVel3D, /
) -> type[CartesianVel3D]:
    """Return the corresponding Cartesian class."""
    return CartesianVel3D


@dispatch
def cartesian_vector_type(
    obj: type[AbstractAcc3D] | AbstractAcc3D, /
) -> type[CartesianAcc3D]:
    """Return the corresponding Cartesian class."""
    return CartesianAcc3D


###############################################################################
# Corresponding time derivative classes

# -----------------------------------------------
# Position -> Velocity


@dispatch
def time_derivative_vector_type(
    obj: type[CartesianPos3D] | CartesianPos3D, /
) -> type[CartesianVel3D]:
    """Return the corresponding time derivative class."""
    return CartesianVel3D


@dispatch
def time_derivative_vector_type(
    obj: type[CylindricalPos] | CylindricalPos, /
) -> type[CylindricalVel]:
    """Return the corresponding time derivative class."""
    return CylindricalVel


@dispatch
def time_derivative_vector_type(
    obj: type[SphericalPos] | SphericalPos, /
) -> type[SphericalVel]:
    """Return the corresponding time derivative class."""
    return SphericalVel


@dispatch
def time_derivative_vector_type(
    obj: type[MathSphericalPos] | MathSphericalPos, /
) -> type[MathSphericalVel]:
    """Return the corresponding time derivative class."""
    return MathSphericalVel


@dispatch
def time_derivative_vector_type(
    obj: type[LonLatSphericalPos] | LonLatSphericalPos, /
) -> type[LonLatSphericalVel]:
    """Return the corresponding time derivative class."""
    return LonLatSphericalVel


@dispatch
def time_derivative_vector_type(
    obj: type[ProlateSpheroidalPos] | ProlateSpheroidalPos, /
) -> type[ProlateSpheroidalVel]:
    """Return the corresponding time derivative class."""
    return ProlateSpheroidalVel


# -----------------------------------------------
# Velocity -> Position


@dispatch
def time_antiderivative_vector_type(
    obj: type[CartesianVel3D] | CartesianVel3D, /
) -> type[CartesianPos3D]:
    """Return the corresponding time antiderivative class."""
    return CartesianPos3D


@dispatch
def time_antiderivative_vector_type(
    obj: type[CylindricalVel] | CylindricalVel, /
) -> type[CylindricalPos]:
    """Return the corresponding time antiderivative class."""
    return CylindricalPos


@dispatch
def time_antiderivative_vector_type(
    obj: type[SphericalVel] | SphericalVel, /
) -> type[SphericalPos]:
    """Return the corresponding time antiderivative class."""
    return SphericalPos


@dispatch
def time_antiderivative_vector_type(
    obj: type[MathSphericalVel] | MathSphericalVel, /
) -> type[MathSphericalPos]:
    """Return the corresponding time derivative class."""
    return MathSphericalPos


@dispatch
def time_antiderivative_vector_type(
    obj: type[LonLatSphericalVel] | LonLatSphericalVel, /
) -> type[LonLatSphericalPos]:
    """Return the corresponding time antiderivative class."""
    return LonLatSphericalPos


@dispatch
def time_antiderivative_vector_type(
    obj: type[LonCosLatSphericalVel] | LonCosLatSphericalVel, /
) -> type[LonLatSphericalPos]:
    """Return the corresponding time antiderivative class."""
    return LonLatSphericalPos


@dispatch
def time_antiderivative_vector_type(
    obj: type[ProlateSpheroidalVel] | ProlateSpheroidalVel, /
) -> type[ProlateSpheroidalPos]:
    """Return the corresponding time antiderivative class."""
    return ProlateSpheroidalPos


# -----------------------------------------------
# Velocity -> Acceleration


@dispatch
def time_derivative_vector_type(
    obj: type[CartesianVel3D] | CartesianVel3D, /
) -> type[CartesianAcc3D]:
    """Return the corresponding time derivative class."""
    return CartesianAcc3D


@dispatch
def time_derivative_vector_type(
    obj: type[CylindricalVel] | CylindricalVel, /
) -> type[CylindricalAcc]:
    """Return the corresponding time derivative class."""
    return CylindricalAcc


@dispatch
def time_derivative_vector_type(
    obj: type[SphericalVel] | SphericalVel, /
) -> type[SphericalAcc]:
    """Return the corresponding time derivative class."""
    return SphericalAcc


@dispatch
def time_derivative_vector_type(
    obj: type[MathSphericalVel] | MathSphericalVel, /
) -> type[MathSphericalAcc]:
    """Return the corresponding time derivative class."""
    return MathSphericalAcc


@dispatch
def time_derivative_vector_type(
    obj: type[LonLatSphericalVel] | LonLatSphericalVel, /
) -> type[LonLatSphericalAcc]:
    """Return the corresponding time derivative class."""
    return LonLatSphericalAcc


@dispatch
def time_derivative_vector_type(
    obj: type[LonCosLatSphericalVel] | LonCosLatSphericalVel, /
) -> type[LonLatSphericalAcc]:
    """Return the corresponding time derivative class."""
    return LonLatSphericalAcc


@dispatch
def time_derivative_vector_type(
    obj: type[ProlateSpheroidalVel] | ProlateSpheroidalVel, /
) -> type[ProlateSpheroidalAcc]:
    """Return the corresponding time derivative class."""
    return ProlateSpheroidalAcc


# -----------------------------------------------
# Acceleration -> Velocity


@dispatch
def time_antiderivative_vector_type(
    obj: type[CartesianAcc3D] | CartesianAcc3D, /
) -> type[CartesianVel3D]:
    """Return the corresponding time antiderivative class."""
    return CartesianVel3D


@dispatch
def time_antiderivative_vector_type(
    obj: type[CylindricalAcc] | CylindricalAcc, /
) -> type[CylindricalVel]:
    """Return the corresponding time antiderivative class."""
    return CylindricalVel


@dispatch
def time_antiderivative_vector_type(
    obj: type[SphericalAcc] | SphericalAcc, /
) -> type[SphericalVel]:
    """Return the corresponding time antiderivative class."""
    return SphericalVel


@dispatch
def time_antiderivative_vector_type(
    obj: type[MathSphericalAcc] | MathSphericalAcc, /
) -> type[MathSphericalVel]:
    """Return the corresponding time antiderivative class."""
    return MathSphericalVel


@dispatch
def time_antiderivative_vector_type(
    obj: type[LonLatSphericalAcc] | LonLatSphericalAcc, /
) -> type[LonLatSphericalVel]:
    """Return the corresponding time antiderivative class."""
    return LonLatSphericalVel


@dispatch
def time_antiderivative_vector_type(
    obj: type[ProlateSpheroidalAcc] | ProlateSpheroidalAcc, /
) -> type[ProlateSpheroidalVel]:
    """Return the corresponding time antiderivative class."""
    return ProlateSpheroidalVel



"""Representation of coordinates in different systems."""

__all__: list[str] = []

from typing import Any, TypeAlias

import jax
from jax import tree as jtu
from jaxtyping import Array, Bool
from quax import quaxify, register

import quaxed.numpy as jnp
import unxt.quantity as uq
from dataclassish import field_items, replace

from .vector import AbstractVector

# ===================================================================

Shape: TypeAlias = tuple[int, ...]


@register(jax.lax.broadcast_in_dim_p)
def broadcast_in_dim_p_absvec(
    operand: AbstractVector, /, *, shape: Shape, **kwargs: Any
) -> AbstractVector:
    """Broadcast in a dimension.

    Examples
    --------
    >>> import quaxed.numpy as jnp
    >>> import coordinax as cx

    Cartesian 1D position, velocity, and acceleration:

    >>> q = cx.vecs.CartesianPos1D.from_([1], "m")
    >>> q.x
    Quantity['length'](Array(1, dtype=int32), unit='m')

    >>> jnp.broadcast_to(q, (1, 1)).x
    Quantity['length'](Array([1], dtype=int32), unit='m')

    >>> p = cx.vecs.CartesianVel1D.from_([1], "m/s")
    >>> p.x
    Quantity['speed'](Array(1, dtype=int32), unit='m / s')

    >>> jnp.broadcast_to(p, (1, 1)).x
    Quantity['speed'](Array([1], dtype=int32), unit='m / s')

    >>> a = cx.vecs.CartesianAcc1D.from_([1], "m/s2")
    >>> a.x
     Quantity['acceleration'](Array(1, dtype=int32), unit='m / s2')

    >>> jnp.broadcast_to(a, (1, 1)).x
    Quantity['acceleration'](Array([1], dtype=int32), unit='m / s2')


    Radial 1D position, velocity, and acceleration:

    >>> q = cx.vecs.RadialPos.from_([1], "m")
    >>> q.r
    Distance(Array(1, dtype=int32), unit='m')

    >>> jnp.broadcast_to(q, (1, 1)).r
    Distance(Array([1], dtype=int32), unit='m')

    >>> p = cx.vecs.RadialVel.from_([1], "m/s")
    >>> p.r
    Quantity['speed'](Array(1, dtype=int32), unit='m / s')

    >>> jnp.broadcast_to(p, (1, 1)).r
    Quantity['speed'](Array([1], dtype=int32), unit='m / s')

    >>> a = cx.vecs.RadialAcc.from_([1], "m/s2")
    >>> a.r
    Quantity['acceleration'](Array(1, dtype=int32), unit='m / s2')

    >>> jnp.broadcast_to(a, (1, 1)).r
    Quantity['acceleration'](Array([1], dtype=int32), unit='m / s2')


    Cartesian 2D position, velocity, and acceleration:

    >>> q = cx.vecs.CartesianPos2D.from_([1, 2], "m")
    >>> print(q)
    <CartesianPos2D (x[m], y[m])
        [1 2]>

    >>> print(jnp.broadcast_to(q, (1, 2)))
    <CartesianPos2D (x[m], y[m])
        [[1 2]]>

    >>> p = cx.vecs.CartesianVel2D.from_([1, 2], "m/s")
    >>> print(p)
    <CartesianVel2D (x[m / s], y[m / s])
        [1 2]>

    >>> print(jnp.broadcast_to(p, (1, 2)))
    <CartesianVel2D (x[m / s], y[m / s])
        [[1 2]]>

    >>> a = cx.vecs.CartesianAcc2D.from_([1, 2], "m/s2")
    >>> print(a)
    <CartesianAcc2D (x[m / s2], y[m / s2])
        [1 2]>

    >>> print(jnp.broadcast_to(a, (1, 2)))
    <CartesianAcc2D (x[m / s2], y[m / s2])
        [[1 2]]>

    Cartesian 3D position, velocity, and acceleration:

    >>> q = cx.CartesianPos3D.from_([1, 2, 3], "m")
    >>> q.x
    Quantity['length'](Array(1, dtype=int32), unit='m')

    >>> jnp.broadcast_to(q, (1, 3)).x
    Quantity['length'](Array([1], dtype=int32), unit='m')

    >>> p = cx.CartesianVel3D.from_([1, 2, 3], "m/s")
    >>> p.x
    Quantity['speed'](Array(1, dtype=int32), unit='m / s')

    >>> jnp.broadcast_to(p, (1, 3)).x
    Quantity['speed'](Array([1], dtype=int32), unit='m / s')

    >>> a = cx.vecs.CartesianAcc3D.from_([1, 2, 3], "m/s2")
    >>> print(a)
    <CartesianAcc3D (x[m / s2], y[m / s2], z[m / s2])
        [1 2 3]>

    >>> print(jnp.broadcast_to(a, (1, 3)))
    <CartesianAcc3D (x[m / s2], y[m / s2], z[m / s2])
        [[1 2 3]]>

    """
    # TODO: use `jax.lax.broadcast_in_dim_p`
    c_shape = shape[:-1]
    return replace(
        operand,
        **{k: jnp.broadcast_to(v, c_shape) for k, v in field_items(operand)},
    )


# ===================================================================


@register(jax.lax.convert_element_type_p)
def convert_element_type_p_absvec(
    operand: AbstractVector, /, **kwargs: Any
) -> AbstractVector:
    """Convert the element type of a quantity.

    Examples
    --------
    >>> import quaxed.lax as qlax
    >>> import coordinax as cx

    >>> vec = cx.vecs.CartesianPosND.from_([1, 2, 3], "m")
    >>> qlax.convert_element_type(vec, float)
    CartesianPosND(q=Quantity[...](value=f32[3], unit=Unit("m")))

    """
    convert_p = quaxify(jax.lax.convert_element_type_p.bind)
    return replace(
        operand,
        **{k: convert_p(v, **kwargs) for k, v in field_items(operand)},
    )


# ===================================================================


@register(jax.lax.eq_p)
def eq_p_absvecs(lhs: AbstractVector, rhs: AbstractVector, /) -> Bool[Array, "..."]:
    """Element-wise equality of two vectors.

    See `AbstractVector.__eq__` for examples.

    """
    # Map the equality over the leaves, which are Quantities.
    comp_tree = jtu.map(
        jnp.equal,
        jtu.leaves(lhs, is_leaf=uq.is_any_quantity),
        jtu.leaves(rhs, is_leaf=uq.is_any_quantity),
        is_leaf=uq.is_any_quantity,
    )

    # Reduce the equality over the leaves.
    return jax.tree.reduce(jnp.logical_and, comp_tree)



"""Transformations between representations."""

__all__: list[str] = []

import functools as ft
from warnings import warn

import jax
from plum import dispatch

import quaxed.numpy as jnp

import coordinax._src.vectors.custom_types as ct
from coordinax._src.vectors import d1, d2, d3
from coordinax._src.vectors.base_pos import AbstractPos
from coordinax._src.vectors.exceptions import IrreversibleDimensionChange
from coordinax._src.vectors.private_api import wrap_vconvert_impl_params

# =============================================================================
# CartesianPos3D


# -----------------------------------------------
# 1D


@dispatch
@wrap_vconvert_impl_params
def vconvert(
    to_vector: type[d1.RadialPos],
    from_vector: type[d3.CartesianPos3D],
    p: ct.ParamsDict,
    /,
    *,
    in_aux: ct.OptAuxDict = None,
    out_aux: ct.OptAuxDict = None,
    units: ct.OptUSys = None,
) -> tuple[ct.ParamsDict, ct.AuxDict]:
    """CartesianPos3D -> RadialPos.

    Examples
    --------
    >>> import warnings
    >>> import coordinax.vecs as cxv

    >>> params = {"x": 1, "y": 2, "z": 3}

    >>> with warnings.catch_warnings():
    ...     warnings.simplefilter("ignore")
    ...     cxv.vconvert(cxv.RadialPos, cxv.CartesianPos3D, params)
    ({'r': Array(3.7416575, dtype=float32, ...)}, {})

    >>> x = cxv.CartesianPos3D.from_([1.0, 2.0, 3.0], "km")
    >>> with warnings.catch_warnings():
    ...     warnings.simplefilter("ignore")
    ...     y = cxv.vconvert(cxv.RadialPos, x)
    >>> print(y)
    <RadialPos (r[km])
        [3.742]>

    """
    warn("irreversible dimension change", IrreversibleDimensionChange, stacklevel=2)
    r = jnp.sqrt(p["x"] ** 2 + p["y"] ** 2 + p["z"] ** 2)
    return {"r": r}, (out_aux or {})


# -----------------------------------------------
# 2D


@dispatch
@wrap_vconvert_impl_params
def vconvert(
    to_vector: type[d2.PolarPos],
    from_vector: type[d3.CartesianPos3D],
    p: ct.ParamsDict,
    /,
    *,
    in_aux: ct.OptAuxDict = None,
    out_aux: ct.OptAuxDict = None,
    units: ct.OptUSys = None,
) -> tuple[ct.ParamsDict, ct.AuxDict]:
    """CartesianPos3D -> Cartesian2D -> AbstractPos2D.

    Examples
    --------
    >>> import warnings
    >>> import coordinax as cx

    >>> params = {"x": 1, "y": 2, "z": 3}
    >>> with warnings.catch_warnings():
    ...     warnings.simplefilter("ignore")
    ...     cx.vecs.vconvert(cx.vecs.PolarPos, cx.vecs.CartesianPos3D, params)
    ({'r': Array(2.236068, dtype=float32, ...),
      'phi': Array(1.1071488, dtype=float32, ...)},
     {})

    >>> x = cx.CartesianPos3D.from_([1.0, 2.0, 3.0], "km")
    >>> with warnings.catch_warnings():
    ...     warnings.simplefilter("ignore")
    ...     y = cx.vconvert(cx.vecs.PolarPos, x)
    >>> print(y)
    <PolarPos (r[km], phi[rad])
        [2.236 1.107]>

    """
    warn("irreversible dimension change", IrreversibleDimensionChange, stacklevel=2)
    r = jnp.hypot(p["x"], p["y"])
    phi = jnp.arctan2(p["y"], p["x"])
    return {"r": r, "phi": phi}, (out_aux or {})


# =============================================================================
# CylindricalPos


# -----------------------------------------------
# 1D


@dispatch
@wrap_vconvert_impl_params
def vconvert(
    to_vector: type[d1.CartesianPos1D],
    from_vector: type[d3.CylindricalPos],
    p: ct.ParamsDict,
    /,
    *,
    in_aux: ct.OptAuxDict = None,
    out_aux: ct.OptAuxDict = None,
    units: ct.OptUSys = None,
) -> tuple[ct.ParamsDict, ct.AuxDict]:
    """CylindricalPos -> CartesianPos1D.

    Examples
    --------
    >>> import warnings
    >>> import unxt as u
    >>> import coordinax.vecs as cxv

    >>> params = {"rho": 1, "phi": 10, "z": 14}
    >>> usys = u.unitsystem("km", "deg")

    >>> with warnings.catch_warnings():
    ...     warnings.simplefilter("ignore")
    ...     cxv.vconvert(cxv.CartesianPos1D, cxv.CylindricalPos,
    ...                       params, units=usys)
    ({'x': Array(0.9848077, dtype=float32, ...)}, {})

    >>> x = cx.vecs.CylindricalPos(rho=u.Quantity(1.0, "km"),
    ...                            phi=u.Quantity(10.0, "deg"),
    ...                            z=u.Quantity(14, "km"))
    >>> with warnings.catch_warnings():
    ...     warnings.simplefilter("ignore")
    ...     y = cx.vconvert(cx.vecs.CartesianPos1D, x)
    >>> print(y)
    <CartesianPos1D (x[km])
        [0.985]>

    """
    warn("irreversible dimension change", IrreversibleDimensionChange, stacklevel=2)
    return {"x": p["rho"] * jnp.cos(p["phi"])}, (out_aux or {})


@dispatch
@wrap_vconvert_impl_params
def vconvert(
    to_vector: type[d1.RadialPos],
    from_vector: type[d3.CylindricalPos],
    p: ct.ParamsDict,
    /,
    *,
    in_aux: ct.OptAuxDict = None,
    out_aux: ct.OptAuxDict = None,
    units: ct.OptUSys = None,
) -> tuple[ct.ParamsDict, ct.AuxDict]:
    """CylindricalPos -> RadialPos.

    Examples
    --------
    >>> import warnings
    >>> import unxt as u
    >>> import coordinax.vecs as cxv

    >>> params = {"rho": 1, "phi": 10, "z": 14}
    >>> usys = u.unitsystem("km", "deg")

    >>> with warnings.catch_warnings():
    ...     warnings.simplefilter("ignore")
    ...     cxv.vconvert(cxv.RadialPos, cxv.CylindricalPos, params, units=usys)
    ({'r': Array(1, dtype=int32, ...)}, {})

    >>> x = cxv.CylindricalPos(rho=u.Quantity(1, "km"),
    ...                        phi=u.Quantity(10, "deg"),
    ...                        z=u.Quantity(14, "km"))
    >>> with warnings.catch_warnings():
    ...     warnings.simplefilter("ignore")
    ...     y = cx.vconvert(cxv.RadialPos, x)
    >>> print(y)
    <RadialPos (r[km])
        [1]>

    """
    warn("irreversible dimension change", IrreversibleDimensionChange, stacklevel=2)
    return {"r": p["rho"]}, (out_aux or {})


# -----------------------------------------------
# 2D


@dispatch
@wrap_vconvert_impl_params
def vconvert(
    to_vector: type[d2.CartesianPos2D],
    from_vector: type[d3.CylindricalPos],
    p: ct.ParamsDict,
    /,
    *,
    in_aux: ct.OptAuxDict = None,
    out_aux: ct.OptAuxDict = None,
    units: ct.OptUSys = None,
) -> tuple[ct.ParamsDict, ct.AuxDict]:
    """CylindricalPos -> CartesianPos2D.

    Examples
    --------
    >>> import warnings
    >>> import unxt as u
    >>> import coordinax.vecs as cxv

    >>> params = {"rho": 1, "phi": 10, "z": 14}
    >>> usys = u.unitsystem("km", "deg")

    >>> with warnings.catch_warnings():
    ...     warnings.simplefilter("ignore")
    ...     cxv.vconvert(cxv.CartesianPos2D, cxv.CylindricalPos,
    ...                       params, units=usys)
    ({'x': Array(0.9848077, dtype=float32, weak_type=True),
      'y': Array(0.17364818, dtype=float32, weak_type=True)},
     {})

    >>> x = cxv.CylindricalPos(rho=u.Quantity(1, "km"),
    ...                        phi=u.Quantity(10, "deg"),
    ...                        z=u.Quantity(14, "km"))
    >>> with warnings.catch_warnings():
    ...     warnings.simplefilter("ignore")
    ...     y = cxv.vconvert(cxv.CartesianPos2D, x)
    >>> print(y)
    <CartesianPos2D (x[km], y[km])
        [0.985 0.174]>

    """
    warn("irreversible dimension change", IrreversibleDimensionChange, stacklevel=2)
    x = p["rho"] * jnp.cos(p["phi"])
    y = p["rho"] * jnp.sin(p["phi"])
    return {"x": x, "y": y}, (out_aux or {})


@dispatch
@wrap_vconvert_impl_params
def vconvert(
    to_vector: type[d2.PolarPos],
    from_vector: type[d3.CylindricalPos],
    p: ct.ParamsDict,
    /,
    *,
    in_aux: ct.OptAuxDict = None,
    out_aux: ct.OptAuxDict = None,
    units: ct.OptUSys = None,
) -> tuple[ct.ParamsDict, ct.AuxDict]:
    """CylindricalPos -> PolarPos.

    Examples
    --------
    >>> import warnings
    >>> import unxt as u
    >>> import coordinax.vecs as cxv

    >>> params = {"rho": 1, "phi": 10, "z": 14}
    >>> usys = u.unitsystem("km", "deg")

    >>> with warnings.catch_warnings():
    ...     warnings.simplefilter("ignore")
    ...     cxv.vconvert(cxv.PolarPos, cxv.CylindricalPos,
    ...                       params, units=usys)
    ({'r': Array(1, dtype=int32, ...),
      'phi': Array(10., dtype=float32, ...)}, {})

    >>> x = cxv.CylindricalPos(rho=u.Quantity(1, "km"),
    ...                        phi=u.Quantity(10, "deg"),
    ...                        z=u.Quantity(14, "km"))
    >>> with warnings.catch_warnings():
    ...     warnings.simplefilter("ignore")
    ...     y = cxv.vconvert(cxv.PolarPos, x)
    >>> print(y)
    <PolarPos (r[km], phi[deg])
        [ 1 10]>

    """
    warn("irreversible dimension change", IrreversibleDimensionChange, stacklevel=2)
    return {"r": p["rho"], "phi": p["phi"]}, (out_aux or {})


# =============================================================================
# SphericalPos


# -----------------------------------------------
# 1D


@dispatch
@wrap_vconvert_impl_params
def vconvert(
    to_vector: type[d1.CartesianPos1D],
    from_vector: type[d3.SphericalPos],
    p: ct.ParamsDict,
    /,
    *,
    in_aux: ct.OptAuxDict = None,
    out_aux: ct.OptAuxDict = None,
    units: ct.OptUSys = None,
) -> tuple[ct.ParamsDict, ct.AuxDict]:
    """SphericalPos -> CartesianPos1D.

    Examples
    --------
    >>> import warnings
    >>> import unxt as u
    >>> import coordinax.vecs as cxv

    >>> params = {"r": 1, "theta": 14, "phi": 10}
    >>> usys = u.unitsystem("km", "deg")

    >>> with warnings.catch_warnings():
    ...     warnings.simplefilter("ignore")
    ...     cxv.vconvert(cxv.CartesianPos1D, cxv.SphericalPos,
    ...                       params, units=usys)
    ({'x': Array(0.23824656, dtype=float32, ...)}, {})

    >>> x = cxv.SphericalPos(r=u.Quantity(1, "km"),
    ...                     theta=u.Quantity(14, "deg"),
    ...                     phi=u.Quantity(10, "deg"))
    >>> with warnings.catch_warnings():
    ...     warnings.simplefilter("ignore")
    ...     y = cxv.vconvert(cxv.CartesianPos1D, x)
    >>> print(y)
    <CartesianPos1D (x[km])
        [0.238]>

    """
    warn("irreversible dimension change", IrreversibleDimensionChange, stacklevel=2)
    x = p["r"] * jnp.sin(p["theta"]) * jnp.cos(p["phi"])
    return {"x": x}, (out_aux or {})


# -----------------------------------------------
# 2D


@dispatch
@wrap_vconvert_impl_params
def vconvert(
    to_vector: type[d2.CartesianPos2D],
    from_vector: type[d3.SphericalPos],
    p: ct.ParamsDict,
    /,
    *,
    in_aux: ct.OptAuxDict = None,
    out_aux: ct.OptAuxDict = None,
    units: ct.OptUSys = None,
) -> tuple[ct.ParamsDict, ct.AuxDict]:
    """SphericalPos -> CartesianPos2D.

    Examples
    --------
    >>> import warnings
    >>> import unxt as u
    >>> import coordinax.vecs as cxv

    >>> params = {"r": 1, "theta": 14, "phi": 10}
    >>> usys = u.unitsystem("km", "deg")

    >>> with warnings.catch_warnings():
    ...     warnings.simplefilter("ignore")
    ...     cxv.vconvert(cxv.CartesianPos2D, cxv.SphericalPos,
    ...                       params, units=usys)
    ({'x': Array(0.23824656, dtype=float32, ...),
      'y': Array(0.0420093, dtype=float32, ...)},
     {})

    >>> x = cxv.SphericalPos(r=u.Quantity(1, "km"),
    ...                     theta=u.Quantity(14, "deg"),
    ...                     phi=u.Quantity(10, "deg"))
    >>> with warnings.catch_warnings():
    ...     warnings.simplefilter("ignore")
    ...     y = cxv.vconvert(cxv.CartesianPos2D, x)
    >>> print(y)
    <CartesianPos2D (x[km], y[km])
        [0.238 0.042]>

    """
    warn("irreversible dimension change", IrreversibleDimensionChange, stacklevel=2)
    x = p["r"] * jnp.sin(p["theta"]) * jnp.cos(p["phi"])
    y = p["r"] * jnp.sin(p["theta"]) * jnp.sin(p["phi"])
    return {"x": x, "y": y}, (out_aux or {})


@dispatch
@wrap_vconvert_impl_params
def vconvert(
    to_vector: type[d2.PolarPos],
    from_vector: type[d3.SphericalPos],
    p: ct.ParamsDict,
    /,
    *,
    in_aux: ct.OptAuxDict = None,
    out_aux: ct.OptAuxDict = None,
    units: ct.OptUSys = None,
) -> tuple[ct.ParamsDict, ct.AuxDict]:
    """SphericalPos -> PolarPos.

    Examples
    --------
    >>> import warnings
    >>> import unxt as u
    >>> import coordinax.vecs as cxv

    >>> params = {"r": 1, "theta": 14, "phi": 10}
    >>> usys = u.unitsystem("km", "deg")

    >>> with warnings.catch_warnings():
    ...     warnings.simplefilter("ignore")
    ...     cxv.vconvert(cxv.PolarPos, cxv.SphericalPos,
    ...                       params, units=usys)
    ({'r': Array(0.2419219, dtype=float32, ...),
     'phi': Array(10., dtype=float32, ...)},
     {})

    >>> x = cxv.SphericalPos(r=u.Quantity(1, "km"),
    ...                     theta=u.Quantity(14, "deg"),
    ...                     phi=u.Quantity(10, "deg"))
    >>> with warnings.catch_warnings():
    ...     warnings.simplefilter("ignore")
    ...     y = cxv.vconvert(cxv.PolarPos, x)
    >>> print(y)
    <PolarPos (r[km], phi[deg])
        [ 0.242 10.   ]>

    """
    warn("irreversible dimension change", IrreversibleDimensionChange, stacklevel=2)
    r = p["r"] * jnp.sin(p["theta"])
    return {"r": r, "phi": p["phi"]}, (out_aux or {})


# =============================================================================
# MathSphericalPos


# -----------------------------------------------
# 1D


@dispatch
@wrap_vconvert_impl_params
def vconvert(
    to_vector: type[d1.CartesianPos1D],
    from_vector: type[d3.MathSphericalPos],
    p: ct.ParamsDict,
    /,
    *,
    in_aux: ct.OptAuxDict = None,
    out_aux: ct.OptAuxDict = None,
    units: ct.OptUSys = None,
) -> tuple[ct.ParamsDict, ct.AuxDict]:
    """MathSphericalPos -> CartesianPos1D.

    Examples
    --------
    >>> import warnings
    >>> import unxt as u
    >>> import coordinax.vecs as cxv

    >>> params = {"r": 1, "theta": 10, "phi": 14}
    >>> usys = u.unitsystem("km", "deg")

    >>> with warnings.catch_warnings():
    ...     warnings.simplefilter("ignore")
    ...     cxv.vconvert(cxv.CartesianPos1D, cxv.MathSphericalPos,
    ...                       params, units=usys)
    ({'x': Array(0.23824656, dtype=float32, ...)}, {})

    >>> x = cx.vecs.MathSphericalPos(r=u.Quantity(1, "km"),
    ...                              theta=u.Quantity(10, "deg"),
    ...                              phi=u.Quantity(14, "deg"))

    >>> with warnings.catch_warnings():
    ...     warnings.simplefilter("ignore")
    ...     y = cx.vconvert(cx.vecs.CartesianPos1D, x)
    >>> print(y)
    <CartesianPos1D (x[km])
        [0.238]>

    """
    warn("irreversible dimension change", IrreversibleDimensionChange, stacklevel=2)
    x = p["r"] * jnp.sin(p["phi"]) * jnp.cos(p["theta"])
    return {"x": x}, (out_aux or {})


# -----------------------------------------------
# 2D


@dispatch
@wrap_vconvert_impl_params
def vconvert(
    to_vector: type[d2.CartesianPos2D],
    from_vector: type[d3.MathSphericalPos],
    p: ct.ParamsDict,
    /,
    *,
    in_aux: ct.OptAuxDict = None,
    out_aux: ct.OptAuxDict = None,
    units: ct.OptUSys = None,
) -> tuple[ct.ParamsDict, ct.AuxDict]:
    """MathSphericalPos -> CartesianPos2D.

    Examples
    --------
    >>> import warnings
    >>> import unxt as u
    >>> import coordinax.vecs as cxv

    >>> params = {"r": 1, "theta": 10, "phi": 14}
    >>> usys = u.unitsystem("km", "deg")

    >>> with warnings.catch_warnings():
    ...     warnings.simplefilter("ignore")
    ...     cxv.vconvert(cxv.CartesianPos2D, cxv.MathSphericalPos,
    ...                       params, units=usys)
    ({'x': Array(0.23824656, dtype=float32, ...),
      'y': Array(0.0420093, dtype=float32, ...)},
     {})

    >>> x = cx.vecs.MathSphericalPos(r=u.Quantity(1, "km"),
    ...                              theta=u.Quantity(10, "deg"),
    ...                              phi=u.Quantity(14, "deg"))

    >>> with warnings.catch_warnings():
    ...     warnings.simplefilter("ignore")
    ...     y = cx.vconvert(cx.vecs.CartesianPos2D, x)
    >>> print(y)
    <CartesianPos2D (x[km], y[km])
        [0.238 0.042]>

    """
    warn("irreversible dimension change", IrreversibleDimensionChange, stacklevel=2)
    x = p["r"] * jnp.sin(p["phi"]) * jnp.cos(p["theta"])
    y = p["r"] * jnp.sin(p["phi"]) * jnp.sin(p["theta"])
    return {"x": x, "y": y}, (out_aux or {})


@dispatch
@wrap_vconvert_impl_params
def vconvert(
    to_vector: type[d2.PolarPos],
    from_vector: type[d3.MathSphericalPos],
    p: ct.ParamsDict,
    /,
    *,
    in_aux: ct.OptAuxDict = None,
    out_aux: ct.OptAuxDict = None,
    units: ct.OptUSys = None,
) -> tuple[ct.ParamsDict, ct.AuxDict]:
    """MathSphericalPos -> PolarPos.

    Examples
    --------
    >>> import warnings
    >>> import unxt as u
    >>> import coordinax.vecs as cxv

    >>> params = {"r": 1, "theta": 10, "phi": 14}
    >>> usys = u.unitsystem("km", "deg")

    >>> with warnings.catch_warnings():
    ...     warnings.simplefilter("ignore")
    ...     cxv.vconvert(cxv.PolarPos, cxv.MathSphericalPos,
    ...                       params, units=usys)
    ({'r': Array(0.2419219, dtype=float32, ...),
      'phi': Array(10., dtype=float32, ...)},
     {})

    >>> x = cxv.MathSphericalPos(r=u.Quantity(1, "km"),
    ...                          theta=u.Quantity(10, "deg"),
    ...                          phi=u.Quantity(14, "deg"))
    >>> with warnings.catch_warnings():
    ...     warnings.simplefilter("ignore")
    ...     y = cxv.vconvert(cxv.PolarPos, x)
    >>> print(y)
    <PolarPos (r[km], phi[deg])
        [ 0.242 10.   ]>

    """
    warn("irreversible dimension change", IrreversibleDimensionChange, stacklevel=2)
    r = p["r"] * jnp.sin(p["phi"])
    return {"r": r, "phi": p["theta"]}, (out_aux or {})


# =============================================================================
# ProlateSpheroidalPos


# -----------------------------------------------
# 1D


@dispatch.multi(
    (type[d1.AbstractPos1D], type[d3.ProlateSpheroidalPos], ct.ParamsDict),
    (type[d2.AbstractPos2D], type[d3.ProlateSpheroidalPos], ct.ParamsDict),
    (type[d3.AbstractPos3D], type[d3.ProlateSpheroidalPos], ct.ParamsDict),
)
@ft.partial(jax.jit, static_argnums=(0, 1), static_argnames=("units",))
def vconvert(
    to_vector: type[AbstractPos],
    from_vector: type[d3.ProlateSpheroidalPos],
    p: ct.ParamsDict,
    /,
    *,
    in_aux: ct.OptAuxDict = None,
    out_aux: ct.OptAuxDict = None,
    units: ct.OptUSys = None,
) -> tuple[ct.ParamsDict, ct.AuxDict]:
    """ProlateSpheroidalPos -> AbstractPos.

    Examples
    --------
    >>> import warnings
    >>> import unxt as u
    >>> import coordinax as cx

    >>> x = cx.vecs.ProlateSpheroidalPos(
    ...     mu=u.Quantity(2, "km2"),
    ...     nu=u.Quantity(0.5, "km2"),
    ...     phi=u.Quantity(0.5, "rad"),
    ...     Delta=u.Quantity(1, "km"),
    ... )

    >>> with warnings.catch_warnings():
    ...     warnings.simplefilter("ignore")
    ...     y = cx.vconvert(cx.vecs.CartesianPos1D, x)
    >>> print(y)
    <CartesianPos1D (x[km])
        [0.621]>

    >>> with warnings.catch_warnings():
    ...     warnings.simplefilter("ignore")
    ...     y = cx.vconvert(cx.vecs.RadialPos, x)
    >>> print(y)
    <RadialPos (r[km])
        [0.707]>

    >>> with warnings.catch_warnings():
    ...     warnings.simplefilter("ignore")
    ...     x2 = cx.vconvert(cx.vecs.CartesianPos2D, x)
    >>> print(x2)
    <CartesianPos2D (x[km], y[km])
        [0.621 0.339]>

    >>> with warnings.catch_warnings():
    ...     warnings.simplefilter("ignore")
    ...     x2 = cx.vconvert(cx.vecs.PolarPos, x)
    >>> print(x2)
    <PolarPos (r[km], phi[rad])
        [0.707 0.5  ]>

    >>> import coordinax.vecs as cxv
    >>> vec = {"mu": 1, "nu": 0.2, "phi": 90}
    >>> in_aux = {"Delta": 0.5}
    >>> usys = u.unitsystem("km", "deg")

    >>> cxv.vconvert(cxv.CylindricalPos, cxv.ProlateSpheroidalPos,
    ...                   vec, in_aux=in_aux, units=usys)
    ({'phi': Array(90., dtype=float32, ...),
      'rho': Array(0.38729832, dtype=float32, ...),
      'z': Array(0.8944272, dtype=float32, ...)},
     {})

    >>> cxv.vconvert(cxv.CartesianPos3D, cxv.ProlateSpheroidalPos,
    ...                   vec, in_aux=in_aux, units=usys)
    ({'x': Array(-1.6929347e-08, dtype=float32, ...),
      'y': Array(0.38729832, dtype=float32, ...),
      'z': Array(0.8944272, dtype=float32, ...)},
     {})

    """
    p, aux = vconvert(
        d3.CylindricalPos, from_vector, p, in_aux=in_aux, out_aux=None, units=units
    )
    p, aux = vconvert(
        to_vector, d3.CylindricalPos, p, in_aux=aux, out_aux=out_aux, units=units
    )
    return p, aux



"""Copyright (c) 2023 coordinax maintainers. All rights reserved."""

__all__ = [
    "vector",
    "vconvert",
    "normalize_vector",
    "cartesian_vector_type",
    "time_derivative_vector_type",
    "time_antiderivative_vector_type",
    "time_nth_derivative_vector_type",
]

from typing import TYPE_CHECKING, Any

from plum import dispatch

if TYPE_CHECKING:
    import coordinax.vecs


@dispatch.abstract
def vector(*args: Any, **kwargs: Any) -> Any:
    """Construct a vector given the arguments."""
    raise NotImplementedError  # pragma: no cover


@dispatch.abstract
def vconvert(target: type[Any], /, *args: Any, **kwargs: Any) -> Any:
    """Transform the current vector to the target vector.

    See the dispatch implementations for more details. Not all transformations
    result in the target vector type, for example
    ``vconvert(type[Cartesian3DPos], FourVector)`` will return a
    `coordinax.vecs.FourVector` with the spatial part in Cartesian coordinates.
    Likewise, `coordinax.vconvert` on `coordinax.Coordinate` instances will
    transform the contained vectors to the target type, returning a
    `coordinax.Coordinate` instance.

    Examples
    --------
    >>> import jax
    >>> import jax.numpy as jnp
    >>> import unxt as u
    >>> import coordinax.vecs as cxv

    ## 1D:

    - Array-valued:

    >>> params = {"x": jnp.array([1.0, 2.0])}
    >>> cxv.vconvert(cxv.RadialPos, cxv.CartesianPos1D, params)
    ({'r': Array([1., 2.], dtype=float32)}, {})

    - Quantity-valued:

    >>> params = {"x": u.Quantity([1.0, 2.0], "m")}
    >>> cxv.vconvert(cxv.RadialPos, cxv.CartesianPos1D, params)
    ({'r': Quantity['length'](Array([1., 2.], dtype=float32), unit='m')},
     {})

    - Vector-valued:

    >>> x = cxv.CartesianPos1D.from_(1, "km")
    >>> y = cxv.vconvert(cxv.RadialPos, x)
    >>> print(y)
    <RadialPos (r[km])
        [1]>

    ## 2D:

    - Array-valued:

    Without unit information "phi" is assumed to be in radians.

    >>> params = {"r": jnp.array([1.0, 2.0]), "phi": jnp.array(3)}
    >>> cxv.vconvert(cxv.CartesianPos2D, cxv.PolarPos, params)
    ({'x': Array([-0.9899925, -1.979985 ], dtype=float32),
      'y': Array([0.14112, 0.28224], dtype=float32)},
     {})

    We can provide that unit information so that "phi" is in degrees:

    >>> usys = u.unitsystem("kpc", "deg")
    >>> cxv.vconvert(cxv.CartesianPos2D, cxv.PolarPos, params, units=usys)
    ({'x': Array([0.9986295, 1.997259 ], dtype=float32),
      'y': Array([0.05233596, 0.10467192], dtype=float32)},
     {})

    - Quantity-valued:

    >>> params = {"r": u.Quantity([1.0, 2.0], "m"), "phi": u.Quantity(3, "deg")}
    >>> cxv.vconvert(cxv.CartesianPos2D, cxv.PolarPos, params)
    ({'x': Quantity[...](Array([0.9986295, 1.997259 ], dtype=float32), unit='m'),
      'y': Quantity[...](Array([0.05233596, 0.10467192], dtype=float32), unit='m')},
     {})

    - Vector-valued:

    >>> x = cxv.CartesianPos2D.from_([3, 4], "km")
    >>> y = cxv.vconvert(cxv.PolarPos, x)
    >>> print(y)
    <PolarPos (r[km], phi[rad])
        [5.    0.927]>

    >>> y = cxv.vconvert(cxv.PolarPos, x, units=u.unitsystem("m", "deg"))
    >>> print(y)
    <PolarPos (r[m], phi[deg])
        [5000.     53.13]>

    ## 3D:

    - Array-valued:

    >>> params = {"x": jnp.array([1.0, 2.0]), "y": jnp.array([3.0, 4.0]),
    ...           "z": jnp.array([5.0, 6.0])}
    >>> params, aux = cxv.vconvert(cxv.SphericalPos, cxv.CartesianPos3D, params)
    >>> jax.tree.map(lambda x: jnp.round(x, 4), params)
    {'phi': Array([1.249 , 1.1071], dtype=float32),
     'r': Array([5.9161   , 7.4832997], dtype=float32),
     'theta': Array([0.5639, 0.6405], dtype=float32)}

    - Quantity-valued:

    >>> params = {"x": u.Quantity([1.0, 2.0], "m"),
    ...           "y": u.Quantity([3.0, 4.0], "m"),
    ...           "z": u.Quantity([5.0, 6.0], "m")}
    >>> params, aux = cxv.vconvert(cxv.SphericalPos, cxv.CartesianPos3D, params)
    >>> jax.tree.map(lambda x: jnp.round(x, 4), params)
    {'phi': Quantity['angle'](Array([1.249 , 1.1071], dtype=float32), unit='rad'),
     'r': Quantity['length'](Array([5.9161   , 7.4832997], dtype=float32), unit='m'),
     'theta': Quantity['angle'](Array([0.5639, 0.6405], dtype=float32), unit='rad')}

    - Vector-valued:

    >>> x = cxv.CartesianPos3D.from_([[1, 3, 5], [2, 4, 6]], "km")
    >>> y = cxv.vconvert(cxv.SphericalPos, x)
    >>> print(y)
    <SphericalPos (r[km], theta[rad], phi[rad])
        [[5.916 0.564 1.249]
         [7.483 0.641 1.107]]>

    """
    raise NotImplementedError  # pragma: no cover


@dispatch.abstract
def normalize_vector(x: Any, /) -> Any:
    """Return the unit vector."""
    raise NotImplementedError  # pragma: no cover


@dispatch.abstract
def cartesian_vector_type(obj: Any, /) -> "type[coordinax.vecs.AbstractVector]":
    """Return the corresponding Cartesian vector type.

    Examples
    --------
    >>> import coordinax as cx

    >>> cx.vecs.cartesian_vector_type(cx.vecs.RadialPos)
    <class 'coordinax...CartesianPos1D'>

    >>> cx.vecs.cartesian_vector_type(cx.vecs.SphericalPos)
    <class 'coordinax...CartesianPos3D'>

    >>> cx.vecs.cartesian_vector_type(cx.vecs.RadialVel)
    <class 'coordinax...CartesianVel1D'>

    >>> cx.vecs.cartesian_vector_type(cx.vecs.TwoSphereAcc)
    <class 'coordinax...CartesianAcc2D'>

    >>> cx.vecs.cartesian_vector_type(cx.vecs.SphericalVel)
    <class 'coordinax...CartesianVel3D'>

    >>> cx.vecs.cartesian_vector_type(cx.vecs.CartesianAcc3D)
    <class 'coordinax...CartesianAcc3D'>

    >>> cx.vecs.cartesian_vector_type(cx.vecs.SphericalAcc)
    <class 'coordinax...CartesianAcc3D'>

    >>> cx.vecs.cartesian_vector_type(cx.vecs.FourVector)
    <class 'coordinax...CartesianPos3D'>

    >>> cx.vecs.cartesian_vector_type(cx.vecs.CartesianPosND)
    <class 'coordinax...CartesianPosND'>

    >>> cx.vecs.cartesian_vector_type(cx.vecs.CartesianVelND)
    <class 'coordinax...CartesianVelND'>

    >>> cx.vecs.cartesian_vector_type(cx.vecs.CartesianAccND)
    <class 'coordinax...CartesianAccND'>

    """
    raise NotImplementedError  # pragma: no cover


@dispatch.abstract
def time_derivative_vector_type(obj: Any, /) -> "type[coordinax.vecs.AbstractVector]":
    """Return the corresponding time derivative vector type.

    Examples
    --------
    >>> import coordinax as cx

    >>> cx.vecs.time_derivative_vector_type(cx.vecs.CartesianPos1D)
    <class 'coordinax...CartesianVel1D'>

    >>> cx.vecs.time_derivative_vector_type(cx.vecs.CartesianVel1D)
    <class 'coordinax...CartesianAcc1D'>

    >>> cx.vecs.time_derivative_vector_type(cx.vecs.RadialPos)
    <class 'coordinax...RadialVel'>

    >>> cx.vecs.time_derivative_vector_type(cx.vecs.RadialVel)
    <class 'coordinax...RadialAcc'>

    >>> cx.vecs.time_derivative_vector_type(cx.vecs.CartesianPos2D)
    <class 'coordinax...CartesianVel2D'>

    >>> cx.vecs.time_derivative_vector_type(cx.vecs.CartesianVel2D)
    <class 'coordinax...CartesianAcc2D'>

    >>> cx.vecs.time_derivative_vector_type(cx.vecs.PolarPos)
    <class 'coordinax...PolarVel'>

    >>> cx.vecs.time_derivative_vector_type(cx.vecs.PolarVel)
    <class 'coordinax...PolarAcc'>

    >>> cx.vecs.time_derivative_vector_type(cx.vecs.CartesianPos3D)
    <class 'coordinax...CartesianVel3D'>

    >>> cx.vecs.time_derivative_vector_type(cx.vecs.CartesianVel3D)
    <class 'coordinax...CartesianAcc3D'>

    >>> cx.vecs.time_derivative_vector_type(cx.vecs.CylindricalPos)
    <class 'coordinax...CylindricalVel'>

    >>> cx.vecs.time_derivative_vector_type(cx.vecs.CylindricalVel)
    <class 'coordinax...CylindricalAcc'>

    >>> cx.vecs.time_derivative_vector_type(cx.vecs.SphericalPos)
    <class 'coordinax...SphericalVel'>

    >>> cx.vecs.time_derivative_vector_type(cx.vecs.SphericalVel)
    <class 'coordinax...SphericalAcc'>

    >>> cx.vecs.time_derivative_vector_type(cx.vecs.MathSphericalPos)
    <class 'coordinax...MathSphericalVel'>

    >>> cx.vecs.time_derivative_vector_type(cx.vecs.MathSphericalVel)
    <class 'coordinax...MathSphericalAcc'>

    >>> cx.vecs.time_derivative_vector_type(cx.vecs.LonLatSphericalPos)
    <class 'coordinax...LonLatSphericalVel'>

    >>> cx.vecs.time_derivative_vector_type(cx.vecs.LonLatSphericalVel)
    <class 'coordinax...LonLatSphericalAcc'>

    >>> cx.vecs.time_derivative_vector_type(cx.vecs.LonCosLatSphericalVel)
    <class 'coordinax...LonLatSphericalAcc'>

    >>> cx.vecs.time_derivative_vector_type(cx.vecs.ProlateSpheroidalPos)
    <class 'coordinax...ProlateSpheroidalVel'>

    >>> cx.vecs.time_derivative_vector_type(cx.vecs.ProlateSpheroidalVel)
    <class 'coordinax...ProlateSpheroidalAcc'>

    >>> cx.vecs.time_derivative_vector_type(cx.vecs.CartesianPosND)
    <class 'coordinax...CartesianVelND'>

    >>> cx.vecs.time_derivative_vector_type(cx.vecs.CartesianVelND)
    <class 'coordinax...CartesianAccND'>

    """
    raise NotImplementedError  # pragma: no cover


@dispatch.abstract
def time_antiderivative_vector_type(
    obj: Any, /
) -> "type[coordinax.vecs.AbstractVector]":
    """Return the corresponding time antiderivative vector type.

    Examples
    --------
    >>> import coordinax as cx

    >>> cx.vecs.time_antiderivative_vector_type(cx.vecs.CartesianVel1D)
    <class 'coordinax...CartesianPos1D'>

    >>> cx.vecs.time_antiderivative_vector_type(cx.vecs.CartesianAcc1D)
    <class 'coordinax...CartesianVel1D'>

    >>> cx.vecs.time_antiderivative_vector_type(cx.vecs.RadialVel)
    <class 'coordinax...RadialPos'>

    >>> cx.vecs.time_antiderivative_vector_type(cx.vecs.RadialAcc)
    <class 'coordinax...RadialVel'>

    >>> cx.vecs.time_antiderivative_vector_type(cx.vecs.CartesianVel2D)
    <class 'coordinax...CartesianPos2D'>

    >>> cx.vecs.time_antiderivative_vector_type(cx.vecs.CartesianAcc2D)
    <class 'coordinax...CartesianVel2D'>

    >>> cx.vecs.time_antiderivative_vector_type(cx.vecs.PolarVel)
    <class 'coordinax...PolarPos'>

    >>> cx.vecs.time_antiderivative_vector_type(cx.vecs.PolarAcc)
    <class 'coordinax...PolarVel'>

    >>> cx.vecs.time_antiderivative_vector_type(cx.vecs.CartesianVel3D)
    <class 'coordinax...CartesianPos3D'>

    >>> cx.vecs.time_antiderivative_vector_type(cx.vecs.CartesianAcc3D)
    <class 'coordinax...CartesianVel3D'>

    >>> cx.vecs.time_antiderivative_vector_type(cx.vecs.CylindricalVel)
    <class 'coordinax...CylindricalPos'>

    >>> cx.vecs.time_antiderivative_vector_type(cx.vecs.CylindricalAcc)
    <class 'coordinax...CylindricalVel'>

    >>> cx.vecs.time_antiderivative_vector_type(cx.vecs.SphericalVel)
    <class 'coordinax...SphericalPos'>

    >>> cx.vecs.time_antiderivative_vector_type(cx.vecs.SphericalAcc)
    <class 'coordinax...SphericalVel'>

    >>> cx.vecs.time_antiderivative_vector_type(cx.vecs.MathSphericalVel)
    <class 'coordinax...MathSphericalPos'>

    >>> cx.vecs.time_antiderivative_vector_type(cx.vecs.MathSphericalAcc)
    <class 'coordinax...MathSphericalVel'>

    >>> cx.vecs.time_antiderivative_vector_type(cx.vecs.LonLatSphericalVel)
    <class 'coordinax...LonLatSphericalPos'>

    >>> cx.vecs.time_antiderivative_vector_type(cx.vecs.LonCosLatSphericalVel)
    <class 'coordinax...LonLatSphericalPos'>

    >>> cx.vecs.time_antiderivative_vector_type(cx.vecs.LonLatSphericalAcc)
    <class 'coordinax...LonLatSphericalVel'>

    >>> cx.vecs.time_antiderivative_vector_type(cx.vecs.ProlateSpheroidalVel)
    <class 'coordinax...ProlateSpheroidalPos'>

    >>> cx.vecs.time_antiderivative_vector_type(cx.vecs.ProlateSpheroidalAcc)
    <class 'coordinax...ProlateSpheroidalVel'>

    >>> cx.vecs.time_antiderivative_vector_type(cx.vecs.CartesianVelND)
    <class 'coordinax...CartesianPosND'>

    >>> cx.vecs.time_antiderivative_vector_type(cx.vecs.CartesianAccND)
    <class 'coordinax...CartesianVelND'>

    """
    raise NotImplementedError  # pragma: no cover


@dispatch.abstract
def time_nth_derivative_vector_type(
    obj: Any, /, *, n: int
) -> "type[coordinax.vecs.AbstractVector]":
    """Return the corresponding time nth derivative vector type.

    Examples
    --------
    >>> import coordinax as cx

    >>> cx.vecs.RadialPos.time_nth_derivative_cls(n=0)
    <class 'coordinax...RadialPos'>

    >>> cx.vecs.RadialPos.time_nth_derivative_cls(n=1)
    <class 'coordinax...RadialVel'>

    >>> cx.vecs.RadialPos.time_nth_derivative_cls(n=2)
    <class 'coordinax...RadialAcc'>

    >>> cx.vecs.RadialVel.time_nth_derivative_cls(n=-1)
    <class 'coordinax...RadialPos'>

    >>> cx.vecs.RadialVel.time_nth_derivative_cls(n=0)
    <class 'coordinax...RadialVel'>

    >>> cx.vecs.RadialVel.time_nth_derivative_cls(n=1)
    <class 'coordinax...RadialAcc'>

    >>> cx.vecs.RadialAcc.time_nth_derivative_cls(n=-2)
    <class 'coordinax...RadialPos'>

    >>> cx.vecs.RadialAcc.time_nth_derivative_cls(n=-1)
    <class 'coordinax...RadialVel'>

    >>> cx.vecs.RadialAcc.time_nth_derivative_cls(n=0)
    <class 'coordinax...RadialAcc'>

    """
    raise NotImplementedError  # pragma: no cover



"""Built-in vector classes."""

__all__ = ["CartesianAccND", "CartesianPosND", "CartesianVelND"]

import functools as ft
from typing import final
from typing_extensions import override

import equinox as eqx
import jax
import quax_blocks

import quaxed.numpy as jnp
import unxt as u

import coordinax._src.custom_types as ct
from .base import AbstractAccND, AbstractPosND, AbstractVelND
from coordinax._src.distances import BBtLength
from coordinax._src.vectors.base.cartesian import AbstractCartesian

##############################################################################
# Position


@final
class CartesianPosND(AbstractPosND, AbstractCartesian, quax_blocks.NumpyNegMixin):
    """N-dimensional Cartesian vector representation.

    Examples
    --------
    >>> import coordinax as cx
    >>> import unxt as u

    A 1D vector:

    >>> q = cx.vecs.CartesianPosND.from_([[1]], "km")
    >>> q.q
    Quantity['length'](Array([[1]], dtype=int32), unit='km')
    >>> q.shape
    (1,)

    A 2D vector:

    >>> q = cx.vecs.CartesianPosND(u.Quantity([1, 2], "km"))
    >>> q.q
    Quantity['length'](Array([1, 2], dtype=int32), unit='km')
    >>> q.shape
    ()

    A 3D vector:

    >>> q = cx.vecs.CartesianPosND(u.Quantity([1, 2, 3], "km"))
    >>> q.q
    Quantity['length'](Array([1, 2, 3], dtype=int32), unit='km')
    >>> q.shape
    ()

    A 4D vector:

    >>> q = cx.vecs.CartesianPosND(u.Quantity([1, 2, 3, 4], "km"))
    >>> q.q
    Quantity['length'](Array([1, 2, 3, 4], dtype=int32), unit='km')
    >>> q.shape
    ()

    A 5D vector:

    >>> q = cx.vecs.CartesianPosND(u.Quantity([1, 2, 3, 4, 5], "km"))
    >>> q.q
    Quantity['length'](Array([1, 2, 3, 4, 5], dtype=int32), unit='km')
    >>> q.shape
    ()

    """

    q: BBtLength = eqx.field(converter=u.Quantity["length"].from_)
    r"""N-D coordinate :math:`\vec{x} \in (-\infty,+\infty)`.

    Should have shape (*batch, F) where F is the number of features /
    dimensions. Arbitrary batch shapes are supported.
    """

    @override
    def _dimensionality(self) -> int:
        """Dimensionality of the vector.

        Examples
        --------
        >>> import coordinax as cx

        A 3D vector:

        >>> cx.vecs.CartesianPosND(u.Quantity([1, 2, 3], "km"))._dimensionality()
        3

        """
        return self.q.shape[-1]

    # ===============================================================
    # Quax API

    @override
    def aval(self) -> jax.core.ShapedArray:
        """Simpler aval than superclass."""
        return self.q.aval()

    # -----------------------------------------------------

    @override
    @ft.partial(eqx.filter_jit, inline=True)
    def norm(self) -> BBtLength:
        """Return the norm of the vector.

        Examples
        --------
        >>> import unxt as u
        >>> import coordinax as cx

        A 3D vector:

        >>> q = cx.vecs.CartesianPosND(u.Quantity([1, 2, 3], "km"))
        >>> q.norm()
        Quantity['length'](Array(3.7416575, dtype=float32), unit='km')

        """
        return jnp.linalg.vector_norm(self.q, axis=-1)


##############################################################################
# Velocity


@final
class CartesianVelND(AbstractCartesian, AbstractVelND):
    """Cartesian differential representation.

    Examples
    --------
    >>> import unxt as u
    >>> import coordinax as cx

    A 1D vector:

    >>> q = cx.vecs.CartesianVelND(u.Quantity([[1]], "km/s"))
    >>> q.q
    Quantity['speed'](Array([[1]], dtype=int32), unit='km / s')
    >>> q.shape
    (1,)

    A 2D vector:

    >>> q = cx.vecs.CartesianVelND(u.Quantity([1, 2], "km/s"))
    >>> q.q
    Quantity['speed'](Array([1, 2], dtype=int32), unit='km / s')
    >>> q.shape
    ()

    A 3D vector:

    >>> q = cx.vecs.CartesianVelND(u.Quantity([1, 2, 3], "km/s"))
    >>> q.q
    Quantity['speed'](Array([1, 2, 3], dtype=int32), unit='km / s')
    >>> q.shape
    ()

    A 4D vector:

    >>> q = cx.vecs.CartesianVelND(u.Quantity([1, 2, 3, 4], "km/s"))
    >>> q.q
    Quantity['speed'](Array([1, 2, 3, 4], dtype=int32), unit='km / s')
    >>> q.shape
    ()

    A 5D vector:

    >>> q = cx.vecs.CartesianVelND(u.Quantity([1, 2, 3, 4, 5], "km/s"))
    >>> q.q
    Quantity['speed'](Array([1, 2, 3, 4, 5], dtype=int32), unit='km / s')
    >>> q.shape
    ()

    """

    q: ct.BBtSpeed = eqx.field(converter=u.Quantity["speed"].from_)
    r"""N-D speed :math:`d\vec{x}/dt \in (-\infty, \infty).

    Should have shape (*batch, F) where F is the number of features /
    dimensions. Arbitrary batch shapes are supported.
    """

    @override
    def _dimensionality(self) -> int:
        """Dimensionality of the vector.

        Examples
        --------
        >>> import coordinax as cx

        A 3D vector:

        >>> cx.vecs.CartesianVelND(u.Quantity([1, 2, 3], "km/s"))._dimensionality()
        3

        """
        return self.q.shape[-1]

    @override
    @ft.partial(eqx.filter_jit, inline=True)
    def norm(self, _: AbstractPosND | None = None, /) -> ct.BBtSpeed:
        """Return the norm of the vector.

        Examples
        --------
        >>> import unxt as u
        >>> import coordinax as cx

        A 3D vector:

        >>> c = cx.vecs.CartesianVelND(u.Quantity([1, 2, 3], "km/s"))
        >>> c.norm()
        Quantity['speed'](Array(3.7416575, dtype=float32), unit='km / s')

        """
        return jnp.linalg.vector_norm(self.q, axis=-1)


##############################################################################
# Acceleration


@final
class CartesianAccND(AbstractCartesian, AbstractAccND):
    """Cartesian N-dimensional acceleration representation.

    Examples
    --------
    >>> import unxt as u
    >>> import coordinax as cx

    A 1D vector:

    >>> q = cx.vecs.CartesianAccND(u.Quantity([[1]], "km/s2"))
    >>> q.q
    Quantity['acceleration'](Array([[1]], dtype=int32), unit='km / s2')
    >>> q.shape
    (1,)

    A 2D vector:

    >>> q = cx.vecs.CartesianAccND(u.Quantity([1, 2], "km/s2"))
    >>> q.q
    Quantity['acceleration'](Array([1, 2], dtype=int32), unit='km / s2')
    >>> q.shape
    ()

    A 3D vector:

    >>> q = cx.vecs.CartesianAccND(u.Quantity([1, 2, 3], "km/s2"))
    >>> q.q
    Quantity['acceleration'](Array([1, 2, 3], dtype=int32), unit='km / s2')
    >>> q.shape
    ()

    A 4D vector:

    >>> q = cx.vecs.CartesianAccND(u.Quantity([1, 2, 3, 4], "km/s2"))
    >>> q.q
    Quantity['acceleration'](Array([1, 2, 3, 4], dtype=int32), unit='km / s2')
    >>> q.shape
    ()

    A 5D vector:

    >>> q = cx.vecs.CartesianAccND(u.Quantity([1, 2, 3, 4, 5], "km/s2"))
    >>> q.q
    Quantity['acceleration'](Array([1, 2, 3, 4, 5], dtype=int32), unit='km / s2')
    >>> q.shape
    ()

    """

    q: ct.BBtAcc = eqx.field(converter=u.Quantity["acceleration"].from_)
    r"""N-D acceleration :math:`d\vec{x}/dt^2 \in (-\infty, \infty).

    Should have shape (*batch, F) where F is the number of features /
    dimensions. Arbitrary batch shapes are supported.
    """

    @override
    def _dimensionality(self) -> int:
        """Dimensionality of the vector.

        Examples
        --------
        >>> import coordinax as cx

        A 3D vector:

        >>> cx.vecs.CartesianAccND(u.Quantity([1, 2, 3], "km/s2"))._dimensionality()
        3

        """
        return self.q.shape[-1]

    @override
    @ft.partial(eqx.filter_jit, inline=True)
    def norm(  # type: ignore[override]
        self,
        velocity: AbstractVelND | None = None,
        position: AbstractPosND | None = None,
        /,
    ) -> ct.BBtAcc:
        """Return the norm of the vector.

        Examples
        --------
        >>> import unxt as u
        >>> import coordinax as cx

        A 3D vector:

        >>> c = cx.vecs.CartesianAccND(u.Quantity([1, 2, 3], "km/s2"))
        >>> c.norm()
        Quantity['acceleration'](Array(3.7416575, dtype=float32), unit='km / s2')

        """
        return jnp.linalg.vector_norm(self.q, axis=-1)



"""Representation of coordinates in different systems."""

__all__: list[str] = []


from plum import dispatch

import coordinax._src.vectors.custom_types as ct
from .base import AbstractAcc1D, AbstractPos1D, AbstractVel1D
from .cartesian import CartesianAcc1D, CartesianPos1D, CartesianVel1D
from .radial import RadialAcc, RadialPos, RadialVel
from coordinax._src.vectors.private_api import combine_aux

###############################################################################
# Vector Transformation

# =============================================================================
# `vconvert_impl`


@dispatch
def vconvert(
    to_vector: type[CartesianPos1D],
    from_vector: type[RadialPos],
    params: ct.ParamsDict,
    /,
    *,
    in_aux: ct.OptAuxDict = None,
    out_aux: ct.OptAuxDict = None,
    units: ct.OptUSys = None,
) -> tuple[ct.ParamsDict, ct.AuxDict]:
    """RadialPos -> CartesianPos1D.

    The `r` coordinate is converted to the `x` coordinate of the 1D system.
    """
    return {"x": params["r"]}, combine_aux(in_aux, out_aux)


@dispatch
def vconvert(
    to_vector: type[RadialPos],
    from_vector: type[CartesianPos1D],
    p: ct.ParamsDict,
    /,
    *,
    in_aux: ct.OptAuxDict = None,
    out_aux: ct.OptAuxDict = None,
    units: ct.OptUSys = None,
) -> tuple[ct.ParamsDict, ct.AuxDict]:
    """CartesianPos1D -> RadialPos.

    The `x` coordinate is converted to the `r` coordinate of the 1D system.
    """
    return {"r": p["x"]}, combine_aux(in_aux, out_aux)


###############################################################################
# Corresponding Cartesian classes


@dispatch
def cartesian_vector_type(
    obj: type[AbstractPos1D] | AbstractPos1D, /
) -> type[AbstractPos1D]:
    """Return the corresponding Cartesian vector class."""
    return CartesianPos1D


@dispatch
def cartesian_vector_type(
    obj: type[AbstractVel1D] | AbstractVel1D, /
) -> type[AbstractVel1D]:
    """Return the corresponding Cartesian vector class."""
    return CartesianVel1D


@dispatch
def cartesian_vector_type(
    obj: type[AbstractAcc1D] | AbstractAcc1D, /
) -> type[AbstractAcc1D]:
    """Return the corresponding Cartesian vector class."""
    return CartesianAcc1D


###############################################################################
# Corresponding time derivative classes

# -----------------------------------------------
# Position -> Velocity


@dispatch
def time_derivative_vector_type(
    obj: type[CartesianPos1D] | CartesianPos1D, /
) -> type[CartesianVel1D]:
    """Return the corresponding time derivative class."""
    return CartesianVel1D


@dispatch
def time_derivative_vector_type(obj: type[RadialPos] | RadialPos, /) -> type[RadialVel]:
    """Return the corresponding time derivative class."""
    return RadialVel


# -----------------------------------------------
# Velocity -> Position


@dispatch
def time_antiderivative_vector_type(
    obj: type[CartesianVel1D] | CartesianVel1D, /
) -> type[CartesianPos1D]:
    """Return the corresponding time antiderivative class."""
    return CartesianPos1D


@dispatch
def time_antiderivative_vector_type(
    obj: type[RadialVel] | RadialVel, /
) -> type[RadialPos]:
    """Return the corresponding time antiderivative class."""
    return RadialPos


# -----------------------------------------------
# Velocity -> Acceleration


@dispatch
def time_derivative_vector_type(
    obj: type[CartesianVel1D] | CartesianVel1D, /
) -> type[CartesianAcc1D]:
    """Return the corresponding time derivative class."""
    return CartesianAcc1D


@dispatch
def time_derivative_vector_type(obj: type[RadialVel] | RadialVel, /) -> type[RadialAcc]:
    """Return the corresponding time derivative class."""
    return RadialAcc


# -----------------------------------------------
# Acceleration -> Velocity


@dispatch
def time_antiderivative_vector_type(
    obj: type[CartesianAcc1D] | CartesianAcc1D, /
) -> type[CartesianVel1D]:
    """Return the corresponding time antiderivative class."""
    return CartesianVel1D


@dispatch
def time_antiderivative_vector_type(
    obj: type[RadialAcc] | RadialAcc, /
) -> type[RadialVel]:
    """Return the corresponding time antiderivative class."""
    return RadialVel



"""Representation of coordinates in different systems."""

__all__ = ["AbstractPos", "POSITION_CLASSES"]

import functools as ft
from inspect import isabstract
from typing import TYPE_CHECKING, Any

import equinox as eqx
import jax
import quax_blocks
from jaxtyping import Array, Shaped
from plum import convert
from quax import quaxify

import quaxed.numpy as jnp
import unxt as u
from unxt.quantity import BareQuantity as FastQ

from coordinax._src.custom_types import BBtScalarQ
from coordinax._src.utils import classproperty
from coordinax._src.vectors import api
from coordinax._src.vectors.base import AbstractVector, ToUnitsOptions
from coordinax._src.vectors.mixins import AvalMixin

if TYPE_CHECKING:
    import coordinax.vecs

_vec_matmul = quaxify(jax.numpy.vectorize(jax.numpy.matmul, signature="(N,N),(N)->(N)"))


class AbstractPos(
    AvalMixin, quax_blocks.NumpyNegMixin["coordinax.vecs.AbstractPos"], AbstractVector
):  # pylint: disable=abstract-method
    """Abstract representation of coordinates in different systems."""

    def __init_subclass__(cls, **kwargs: Any) -> None:
        """Initialize the subclass.

        The subclass is registered if it is not an abstract class.
        """
        # TODO: a more robust check using equinox.
        if isabstract(cls) or cls.__name__.startswith("Abstract"):
            return

        POSITION_CLASSES_MUTABLE[cls] = None

    # ===============================================================
    # Vector API

    @classproperty
    @classmethod
    def cartesian_type(cls) -> "type[coordinax.vecs.AbstractPos]":
        """Return the corresponding Cartesian vector class."""
        return api.cartesian_vector_type(cls)

    @classproperty
    @classmethod
    def time_derivative_cls(cls) -> "type[coordinax.vecs.AbstractVel]":
        """Return the corresponding time derivative class."""
        return api.time_derivative_vector_type(cls)

    @classproperty
    @classmethod
    def time_antiderivative_cls(cls) -> "type[coordinax.vecs.AbstractVector]":
        """Return the corresponding time antiderivative class."""
        return api.time_antiderivative_vector_type(cls)

    @classmethod
    def time_nth_derivative_cls(cls, n: int) -> "type[coordinax.vecs.AbstractVector]":
        """Return the corresponding time nth derivative class."""
        return api.time_nth_derivative_vector_type(cls, n=n)

    # ===============================================================
    # Python API

    def __rmatmul__(self, other: Shaped[Array, "N N"]) -> Any:
        """Matrix multiplication with another object.

        Examples
        --------
        >>> import quaxed.numpy as jnp
        >>> import coordinax as cx

        >>> q = cx.CartesianPos3D.from_([[1., 0, 0 ],
        ...                              [0 , 1, 0 ],
        ...                              [0 , 0, 1 ]], "kpc")[None]
        >>> q.shape
        (1, 3)

        >>> R_z = rot = jnp.asarray([[0.0, -1, 0], [1, 0, 0], [0, 0, 1]])

        >>> print(R_z @ q)
        <CartesianPos3D (x[kpc], y[kpc], z[kpc])
            [[[ 0.  1.  0.]
              [-1.  0.  0.]
              [ 0.  0.  1.]]]>

        """
        # TODO: figure out how to do this without converting back to arrays.
        cart_cls = self.cartesian_type
        cartvec = self.vconvert(cart_cls)
        q: FastQ = convert(cartvec.uconvert(ToUnitsOptions.consistent), FastQ)
        newq = _vec_matmul(other, q)
        newvec = cart_cls.from_(newq)
        return newvec.vconvert(type(self))

    def __abs__(self) -> u.AbstractQuantity:
        """Return the norm of the vector.

        Examples
        --------
        >>> import coordinax as cx
        >>> vec = cx.vecs.CartesianPos2D.from_([3, 4], "m")
        >>> abs(vec)
        BareQuantity(Array(5., dtype=float32), unit='m')

        """
        return self.norm()  # type: ignore[misc]

    # ===============================================================
    # Convenience methods

    @ft.partial(eqx.filter_jit, inline=True)
    def norm(self) -> BBtScalarQ:
        """Return the norm of the vector.

        Returns
        -------
        Quantity
            The norm of the vector.

        Examples
        --------
        >>> import unxt as u
        >>> import coordinax as cx

        >>> v = cx.vecs.CartesianPos1D.from_([-1], "km")
        >>> v.norm()
        BareQuantity(Array(1., dtype=float32), unit='km')

        >>> v = cx.vecs.CartesianPos2D.from_([3, 4], "km")
        >>> v.norm()
        BareQuantity(Array(5., dtype=float32), unit='km')

        >>> v = cx.vecs.PolarPos(r=u.Quantity(3, "km"), phi=u.Quantity(90, "deg"))
        >>> v.norm()
        Distance(Array(3, dtype=int32, ...), unit='km')

        >>> v = cx.CartesianPos3D.from_([1, 2, 3], "m")
        >>> v.norm()
        BareQuantity(Array(3.7416575, dtype=float32), unit='m')

        """
        return jnp.linalg.vector_norm(self, axis=-1)  # type: ignore[arg-type]


#: Registered position classes.
POSITION_CLASSES_MUTABLE: dict[type[AbstractPos], None] = {}
POSITION_CLASSES = POSITION_CLASSES_MUTABLE.keys()



"""Register for `coordinax.ops`."""

__all__: list[str] = []


from dataclassish import replace

from .coordinate import Coordinate
from coordinax._src.operators import AbstractOperator


@AbstractOperator.__call__.dispatch  # type: ignore[misc]
def call(self: AbstractOperator, x: Coordinate, /) -> Coordinate:
    """Dispatch to the operator's `__call__` method.

    Examples
    --------
    >>> import coordinax as cx

    >>> coord = cx.Coordinate(cx.CartesianPos3D.from_([1, 2, 3], "kpc"),
    ...                       cx.frames.ICRS())
    >>> coord
    Coordinate(
        data=Space({ 'length': CartesianPos3D( ... ) }),
        frame=ICRS()
    )

    >>> op = cx.ops.GalileanSpatialTranslation.from_([-1, -1, -1], "kpc")

    >>> new_coord = op(coord)
    >>> print(new_coord.data["length"])
    <CartesianPos3D (x[kpc], y[kpc], z[kpc])
        [0 1 2]>

    """
    return replace(x, data=self(x.data))



"""`coordinax.distance` module."""

__all__ = [
    # Classes
    "AbstractDistance",
    "Distance",
    "DistanceModulus",
    "Parallax",
    # funcs
    "distance",
    "parallax",
    "distance_modulus",
]

from jaxtyping import install_import_hook

from .setup_package import RUNTIME_TYPECHECKER

with install_import_hook("coordinax.distance", RUNTIME_TYPECHECKER):
    from ._src.distances import (
        AbstractDistance,
        Distance,
        DistanceModulus,
        Parallax,
        distance,
        distance_modulus,
        parallax,
    )


del RUNTIME_TYPECHECKER, install_import_hook



"""distance functions."""

__all__ = [
    "distance",
    "parallax",
    "distance_modulus",
]


from typing import Any

from jaxtyping import ArrayLike
from plum import dispatch

import quaxed.numpy as jnp
import unxt as u
from unxt.quantity import BareQuantity

from .measures import Distance, DistanceModulus, Parallax

parallax_base_length = u.Quantity(1, "AU")

#####################################################################
# Distance constructor


@dispatch
def distance(value: ArrayLike, unit: Any, /, **kw: Any) -> Distance:
    """Construct a distance.

    Examples
    --------
    >>> import unxt as u
    >>> import coordinax.distance as cxd

    >>> cxd.distance(1, "kpc")
    Distance(Array(1, dtype=int32, weak_type=True), unit='kpc')

    """
    return Distance(jnp.asarray(value, **kw), unit)


@dispatch
def distance(d: Distance, /, **kw: Any) -> Distance:
    """Compute distance from distance.

    Examples
    --------
    >>> import unxt as u
    >>> import coordinax.distance as cxd

    >>> d = cxd.Distance(1, "kpc")
    >>> cxd.distance(d) is d
    True

    >>> cxd.distance(d, dtype=float)
    Distance(Array(1., dtype=float32), unit='kpc')

    """
    if len(kw) == 0:
        return d
    return jnp.asarray(d, **kw)


@dispatch
def distance(d: u.Quantity["length"], /, **kw: Any) -> Distance:
    """Compute distance from distance.

    Examples
    --------
    >>> import unxt as u
    >>> import coordinax.distance as cxd

    >>> q = u.Quantity(1, "kpc")
    >>> cxd.distance(q, dtype=float)
    Distance(Array(1., dtype=float32), unit='kpc')

    """
    unit = u.unit_of(d)
    return Distance(jnp.asarray(d.ustrip(unit), **kw), unit)


@dispatch
def distance(p: Parallax | u.Quantity["angle"], /, **kw: Any) -> Distance:
    """Compute distance from parallax.

    Examples
    --------
    >>> import unxt as u
    >>> import coordinax.distance as cxd

    >>> p = cxd.Parallax(1, "mas")
    >>> cxd.distance(p).uconvert("pc").round(2)
    Distance(Array(1000., dtype=float32, weak_type=True), unit='pc')

    >>> q = u.Quantity(1, "mas")
    >>> cxd.distance(q).uconvert("pc").round(2)
    Distance(Array(1000., dtype=float32, weak_type=True), unit='pc')

    """
    d = parallax_base_length / jnp.tan(p)  # [AU]
    unit = u.unit_of(d)
    return Distance(jnp.asarray(d.ustrip(unit), **kw), unit)


@dispatch
def distance(dm: DistanceModulus | u.Quantity["mag"], /, **kw: Any) -> Distance:
    """Compute distance from distance modulus.

    Examples
    --------
    >>> import unxt as u
    >>> import coordinax.distance as cxd

    >>> dm = cxd.DistanceModulus(10, "mag")
    >>> cxd.distance(dm).uconvert("pc").round(2)
    Distance(Array(1000., dtype=float32, weak_type=True), unit='pc')

    >>> q = u.Quantity(10, "mag")
    >>> cxd.distance(q).uconvert("pc").round(2)
    Distance(Array(1000., dtype=float32, weak_type=True), unit='pc')

    """
    d = 10 ** (1 + dm.ustrip("mag") / 5)
    return Distance(jnp.asarray(d, **kw), "pc")


#####################################################################
# Parallax constructor


@dispatch
def parallax(value: ArrayLike, unit: Any, /, **kw: Any) -> Parallax:
    """Construct a distance.

    Examples
    --------
    >>> import unxt as u
    >>> import coordinax.distance as cxd

    >>> cxd.parallax(1, "mas")
    Parallax(Array(1, dtype=int32, weak_type=True), unit='mas')

    """
    return Parallax(jnp.asarray(value, **kw), unit)


@dispatch
def parallax(p: Parallax, /, **kw: Any) -> Parallax:
    """Compute parallax from parallax.

    Examples
    --------
    >>> import unxt as u
    >>> import coordinax.distance as cxd

    >>> p = cxd.Parallax(1, "mas")
    >>> cxd.parallax(p) is p
    True

    >>> cxd.parallax(p, dtype=float)
    Parallax(Array(1., dtype=float32), unit='mas')

    """
    if len(kw) == 0:
        return p
    return jnp.asarray(p, **kw)


@dispatch
def parallax(p: u.Quantity["angle"], /, **kw: Any) -> Parallax:
    """Compute parallax from parallax.

    Examples
    --------
    >>> import unxt as u
    >>> import coordinax.distance as cxd

    >>> q = u.Quantity(1, "mas")
    >>> cxd.parallax(q, dtype=float)
    Parallax(Array(1., dtype=float32), unit='mas')

    """
    unit = u.unit_of(p)
    return Parallax(jnp.asarray(p.ustrip(unit), **kw), unit)


@dispatch
def parallax(d: Distance | u.Quantity["length"], /, **kw: Any) -> Parallax:
    """Compute parallax from distance.

    Examples
    --------
    >>> import unxt as u
    >>> import coordinax.distance as cxd

    >>> d = cxd.Distance(10, "pc")
    >>> cxd.parallax(d).uconvert("mas").round(2)
    Parallax(Array(100., dtype=float32, ...), unit='mas')

    >>> q = u.Quantity(10, "pc")
    >>> cxd.parallax(q).uconvert("mas").round(2)
    Parallax(Array(100., dtype=float32, ...), unit='mas')

    """
    p = jnp.atan2(parallax_base_length, d)
    return Parallax(jnp.asarray(p.value, **kw), p.unit)


@dispatch
def parallax(dm: DistanceModulus | u.Quantity["mag"], /, **kw: Any) -> Parallax:
    """Convert distance modulus to parallax.

    Examples
    --------
    >>> import unxt as u
    >>> import coordinax.distance as cxd

    >>> dm = cxd.DistanceModulus(10, "mag")
    >>> cxd.parallax(dm).uconvert("mas").round(2)
    Parallax(Array(1., dtype=float32, weak_type=True), unit='mas')

    """
    d = BareQuantity(10 ** (1 + dm.ustrip("mag") / 5), "pc")
    p = jnp.atan2(parallax_base_length, d)
    unit = u.unit_of(p)
    return Parallax(jnp.asarray(p.ustrip(unit), **kw), unit)


#####################################################################
# Distance Modulus constructor


@dispatch
def distance_modulus(value: ArrayLike, unit: Any, /, **kw: Any) -> DistanceModulus:
    """Construct a distance.

    Examples
    --------
    >>> import unxt as u
    >>> import coordinax.distance as cxd

    >>> cxd.distance_modulus(1, "mag")
    DistanceModulus(Array(1, dtype=int32, weak_type=True), unit='mag')

    """
    return DistanceModulus(jnp.asarray(value, **kw), unit)


@dispatch
def distance_modulus(dm: DistanceModulus, /, **kw: Any) -> DistanceModulus:
    """Compute distance modulus from distance modulus.

    Examples
    --------
    >>> import unxt as u
    >>> import coordinax.distance as cxd

    >>> dm = cxd.DistanceModulus(1, "mag")
    >>> cxd.distance_modulus(dm) is dm
    True

    >>> cxd.distance_modulus(dm, dtype=float)
    DistanceModulus(Array(1., dtype=float32), unit='mag')

    """
    if len(kw) == 0:
        return dm
    return jnp.asarray(dm, **kw)


@dispatch
def distance_modulus(dm: u.Quantity["mag"], /, **kw: Any) -> DistanceModulus:
    """Compute parallax from parallax.

    Examples
    --------
    >>> import unxt as u
    >>> import coordinax.distance as cxd

    >>> q = u.Quantity(1, "mag")
    >>> cxd.distance_modulus(q)
    DistanceModulus(Array(1, dtype=int32, weak_type=True), unit='mag')

    """
    unit = u.unit_of(dm)
    return DistanceModulus(jnp.asarray(u.ustrip(unit, dm), **kw), unit)


@dispatch
def distance_modulus(
    d: Distance | u.Quantity["length"], /, **kw: Any
) -> DistanceModulus:
    """Compute distance modulus from distance.

    Examples
    --------
    >>> import unxt as u
    >>> import coordinax.distance as cxd

    >>> d = cxd.Distance(1, "pc")
    >>> cxd.distance_modulus(d)
    DistanceModulus(Array(-5., dtype=float32), unit='mag')

    >>> q = u.Quantity(1, "pc")
    >>> cxd.distance_modulus(q)
    DistanceModulus(Array(-5., dtype=float32), unit='mag')

    """
    dm = 5 * jnp.log10(d.ustrip("pc")) - 5
    return DistanceModulus(jnp.asarray(dm, **kw), "mag")


@dispatch
def distance_modulus(
    p: Parallax | u.Quantity["angle"], /, **kw: Any
) -> DistanceModulus:
    """Compute distance modulus from parallax.

    Examples
    --------
    >>> import unxt as u
    >>> import coordinax.distance as cxd

    >>> p = cxd.Parallax(1, "mas")
    >>> cxd.distance_modulus(p)
    DistanceModulus(Array(10., dtype=float32), unit='mag')

    >>> q = u.Quantity(1, "mas")
    >>> cxd.distance_modulus(q)
    DistanceModulus(Array(10., dtype=float32), unit='mag')

    """
    d = parallax_base_length / jnp.tan(p)  # [AU]
    dm = 5 * jnp.log10(d.ustrip("pc")) - 5
    return DistanceModulus(jnp.asarray(dm, **kw), "mag")



"""Representation of coordinates in different systems."""

__all__: list[str] = []

from dataclasses import replace
from typing import Any, cast

import equinox as eqx
import jax
from jaxtyping import ArrayLike
from quax import register

import quaxed.lax as qlax
import unxt as u

from .cartesian import CartesianPosND
from coordinax._src.vectors.base_pos import AbstractPos

# ------------------------------------------------


@register(jax.lax.add_p)
def add_p_cartnd_pos(lhs: CartesianPosND, rhs: AbstractPos, /) -> CartesianPosND:
    """Add two vectors.

    Examples
    --------
    >>> import coordinax as cx

    A 3D vector:

    >>> q1 = cx.vecs.CartesianPosND.from_([1, 2, 3], "km")
    >>> q2 = cx.vecs.CartesianPosND.from_([2, 3, 4], "km")
    >>> (q1 + q2).q
    Quantity['length'](Array([3, 5, 7], dtype=int32), unit='km')

    """
    cart = cast(CartesianPosND, rhs.vconvert(CartesianPosND))
    return replace(lhs, q=lhs.q + cart.q)


# ------------------------------------------------
# Dot product
# TODO: see implementation in https://github.com/google/tree-math for how to do
# this more generally.


@register(jax.lax.dot_general_p)
def dot_general_p_cartnds(
    lhs: CartesianPosND, rhs: CartesianPosND, /, **kwargs: Any
) -> u.AbstractQuantity:
    """Dot product of two vectors.

    Examples
    --------
    >>> import quaxed.numpy as jnp
    >>> import unxt as u
    >>> import coordinax as cx

    >>> q1 = cx.vecs.CartesianPosND.from_([1, 2, 3], "m")
    >>> q2 = cx.vecs.CartesianPosND.from_([4, 5, 6], "m")

    >>> jnp.dot(q1, q2)
    Quantity['area'](Array(32, dtype=int32), unit='m2')

    """
    return qlax.dot_general(lhs.q, rhs.q, **kwargs)


# ------------------------------------------------


@register(jax.lax.mul_p)
def mul_p_arraylike_cartnd(lhs: ArrayLike, rhs: CartesianPosND, /) -> CartesianPosND:
    """Scale a position by a scalar.

    Examples
    --------
    >>> import quaxed.numpy as jnp
    >>> import unxt as u
    >>> import coordinax as cx

    >>> v = cx.vecs.CartesianPosND(u.Quantity([1, 2, 3, 4, 5], "km"))
    >>> jnp.multiply(2, v).q
    Quantity['length'](Array([ 2,  4,  6,  8, 10], dtype=int32), unit='km')

    """
    # Validation
    lhs = eqx.error_if(
        lhs, any(jax.numpy.shape(lhs)), f"must be a scalar, not {type(lhs)}"
    )

    # Scale the components
    return replace(rhs, q=lhs * rhs.q)


@register(jax.lax.neg_p)
def neg_p_cartnd(obj: CartesianPosND, /) -> CartesianPosND:
    """Negate the `coordinax.CartesianPosND`.

    Examples
    --------
    >>> import unxt as u
    >>> import coordinax as cx

    A 3D vector:

    >>> vec = cx.vecs.CartesianPosND(u.Quantity([1, 2, 3], "km"))
    >>> (-vec).q
    Quantity['length'](Array([-1, -2, -3], dtype=int32), unit='km')

    """
    return jax.tree.map(qlax.neg, obj)


@register(jax.lax.sub_p)
def sub_p_cartnd_pos(lhs: CartesianPosND, rhs: AbstractPos, /) -> CartesianPosND:
    """Subtract two vectors.

    Examples
    --------
    >>> import unxt as u
    >>> import coordinax as cx

    A 3D vector:

    >>> q1 = cx.vecs.CartesianPosND(u.Quantity([1, 2, 3], "km"))
    >>> q2 = cx.vecs.CartesianPosND(u.Quantity([2, 3, 4], "km"))
    >>> print(q1 - q2)
    <CartesianPosND (q[km])
        [[-1]
         [-1]
         [-1]]>

    """
    cart = cast(CartesianPosND, rhs.vconvert(CartesianPosND))
    return replace(lhs, q=lhs.q - cart.q)
